#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import copy
import dill
import numpy as np
import numpy.linalg as npla
import scipy.optimize as sciopt

import SpectralToolbox.Spectral1D as S1D

from .AdaptivityBase import Builder
from TransportMaps.Distributions import \
    PushForwardTransportMapDistribution, PullBackTransportMapDistribution
from TransportMaps.Routines import \
    laplace_approximation, grad_x_grad_t_kl_divergence, grad_t_kl_divergence
from TransportMaps.Misc import generate_total_order_midxs, \
    mpi_map, mpi_alloc_dmem, mpi_map_alloc_dmem, \
    ExpectationReduce, AbsExpectationReduce
from TransportMaps.Defaults import Default_LinearSpanTriangularTransportMap
from TransportMaps.Diagnostics.Routines import variance_approx_kl
from TransportMaps.Maps.FullTransportMaps import LinearTransportMap
from TransportMaps.Functionals.LinearSpanApproximationBase import \
    LinearSpanApproximation

__all__ = ['KullbackLeiblerBuilder',
           'SequentialKullbackLeiblerBuilder',
           'ToleranceSequentialKullbackLeiblerBuilder',
           'FirstVariationKullbackLeiblerBuilder']

nax = np.newaxis

class KullbackLeiblerBuilder(Builder):
    r""" Basis builder through minimization of kl divergence

    Given distribution :math:`\nu_\rho` and :math:`\nu_\pi`,
    and the parametric transport map :math:`T[{\bf a}]`,
    provides the functionalities to solve the problem

    .. math::

       \arg\min_{\bf a}\mathcal{D}_{\rm KL}\left(
       T[{\bf a}]_\sharp\rho \Vert \pi\right)

    up to a chosen tolerance.

    Args:
      base_distribution (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        distribution :math:`\nu_\rho`
      target_distribution (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        distribution :math:`\nu_\pi`
      transport_map (:class:`TransportMap<TransportMaps.Maps.TransportMap>`):
        transport map :math:`T`
      solve_params (dict): dictionary of parameters for solution
      validator (:class:`Validator<TransportMaps.Diagnostic.Validator>`):
        validator to be used to check stability of the solution
      callback (function): function taking a map and optional additional arguments
        which is called whenever it is deemed necessary by the chosen algorithm
        (e.g. for storing purposes)
      callback_kwargs (dict): additional arguments to be provided to the function
        ``callback``.
      verbosity (int): level of verbosity of the builder
    """
    def __init__(self, base_distribution, target_distribution,
                 transport_map, solve_params, validator=None,
                 callback=None, callback_kwargs={}, verbosity=0):
        self.base_distribution = base_distribution
        self.target_distribution = target_distribution
        self.transport_map = transport_map
        self.solve_params = solve_params
        self.validator = validator
        self.callback = callback
        self.callback_kwargs = callback_kwargs
        self.verbosity = verbosity
        super(KullbackLeiblerBuilder, self).__init__()

    def __getstate__(self):
        out = super(KullbackLeiblerBuilder, self).__getstate__()
        out.pop('callback', None)
        out.pop('callback_kwargs', None)
        return out

    def __setstate__(self, state):
        super(KullbackLeiblerBuilder, self).__setstate__(state)

    def set_mpi_pool(self, mpi_pool):
        self.solve_params['mpi_pool'] = mpi_pool
    
    def solve(self, reloading=False):
        r"""          
        Returns:
          (:class:`TransportMaps.Maps.TransportMap`) -- the transport map fitted.
        """
        if reloading:
            self.solve_params['x0'] = self.transport_map.coeffs
        if 'x0' not in self.solve_params or self.solve_params['x0'] is None:
            self.solve_params['x0'] = \
                self.transport_map.get_default_init_values_minimize_kl_divergence()
        if self.validator is None:
            pull_tar = PullBackTransportMapDistribution(
                self.transport_map, self.target_distribution)
            log = self.transport_map.minimize_kl_divergence(
                self.base_distribution, pull_tar, **self.solve_params)
            if not log['success']:
                self.transport_map.coeffs = self.solve_params['x0']
        else:
            log = self.validator.solve_to_tolerance(
                self.transport_map,
                self.base_distribution, self.target_distribution,
                self.solve_params)
        return self.transport_map, log

class SequentialKullbackLeiblerBuilder(KullbackLeiblerBuilder):
    r""" Solve over a list of maps, using the former to warm start the next one

    Given distribution :math:`\nu_\rho` and :math:`\nu_\pi`,
    and the list of parametric transport maps
    :math:`[T_1[{\bf a}_1,\ldots,T_n[{\bf a}_n]`,
    provides the functionalities to solve the problems

    .. math::

       \arg\min_{{\bf a}_i}\mathcal{D}_{\rm KL}\left(
       T_i[{\bf a}_i]_\sharp\rho \Vert \pi\right)

    up to a chosen tolerance, where the numerical solution for map
    :math:`T_{i+1}` is started at :math:`T_i`

    Args:
      base_distribution (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        distribution :math:`\nu_\rho`
      target_distribution (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        distribution :math:`\nu_\pi`
      transport_map_list (:class:`list` of :class:`TransportMap<TransportMaps.Maps.TransportMap>`):
        list of transports map :math:`T_i`
      solve_params_list (:class:`list` of :class:`dict`):
        list of dictionaries of parameters for solution
      validator (:class:`Validator<TransportMaps.Diagnostic.Validator>`):
        validator to be used to check stability of the solution
      regression_params_list (:class:`list` of :class:`dict`):
        list of dictionaries of parameters for the regression between :math:`T_i` and
        :math:`T_{i+1}`
      verbosity (int): level of verbosity of the builder
    """
    def __init__(self, base_distribution, target_distribution,
                 transport_map_list, solve_params_list,
                 validator=None, regression_params_list=None,
                 callback=None, callback_kwargs={}, verbosity=0):
        self.solve_counter = 0
        self.transport_map_list = transport_map_list
        self.solve_params_list = solve_params_list
        self.regression_params_list = regression_params_list
        super(SequentialKullbackLeiblerBuilder,
              self).__init__(
                  base_distribution, target_distribution,
                  transport_map_list[0], solve_params_list[0], validator,
                  callback=callback, callback_kwargs=callback_kwargs,
                  verbosity=verbosity)

    def set_mpi_pool(self, mpi_pool):
        for solve_params in self.solve_params_list:
            solve_params['mpi_pool'] = mpi_pool
        
    def solve(self, reloading=False):
        r"""
        Returns:
          (:class:`TransportMaps.Maps.TransportMap`) -- the last transport map fitted.
        """
        if not reloading:
            self.solve_counter = 0
        if self.solve_counter == 0:
            self.transport_map = self.transport_map_list[0]
            self.solve_params = self.solve_params_list[0]
            tm, log = super(SequentialKullbackLeiblerBuilder, self).solve()
            if not log['success']:
                tm.coeffs = x0
                return tm, log
            self.solve_counter += 1
        tm_old = self.transport_map_list[self.solve_counter-1]
        for self.transport_map, self.solve_params in zip(
                self.transport_map_list[self.solve_counter:],
                self.solve_params_list[self.solve_counter:]):
            # Here we are assuming nested basis
            for c1, c2 in zip(tm_old.approx_list, self.transport_map.approx_list):
                # Constant part
                for i1, midx1 in enumerate(c1.c.multi_idxs):
                    for i2, midx2 in enumerate(c2.c.multi_idxs):
                        if midx1 == midx2:
                            break
                    c2.c.coeffs[i2] = c1.c.coeffs[i1]
                # Integrated part
                for i1, midx1 in enumerate(c1.h.multi_idxs):
                    for i2, midx2 in enumerate(c2.h.multi_idxs):
                        if midx1 == midx2:
                            break
                    c2.h.coeffs[i2] = c1.h.coeffs[i1]
            # solve for the new map using regressed starting point
            self.solve_params['x0'] = self.transport_map.coeffs
            tm, log = super(SequentialKullbackLeiblerBuilder, self).solve()
            if not log['success']:
                return tm_old, log
            tm_old = tm
            self.solve_counter += 1
        return tm, log

class ToleranceSequentialKullbackLeiblerBuilder(KullbackLeiblerBuilder):
    r""" Solve over a list of maps, using the former to warm start the next one, until a target tolerance is met

    Given distribution :math:`\nu_\rho` and :math:`\nu_\pi`,
    and the list of parametric transport maps
    :math:`[T_1[{\bf a}_1,\ldots,T_n[{\bf a}_n]`,
    provides the functionalities to solve the problems

    .. math::

       \arg\min_{{\bf a}_i}\mathcal{D}_{\rm KL}\left(
       T_i[{\bf a}_i]_\sharp\rho \Vert \pi\right)

    up to a chosen tolerance, where the numerical solution for map
    :math:`T_{i+1}` is started at :math:`T_i`

    Args:
      base_distribution (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        distribution :math:`\nu_\rho`
      target_distribution (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        distribution :math:`\nu_\pi`
      transport_map_list (:class:`list` of :class:`TransportMap<TransportMaps.Maps.TransportMap>`):
        list of transports map :math:`T_i`
      solve_params_list (:class:`list` of :class:`dict`):
        list of dictionaries of parameters for solution
      regression_params_list (:class:`list` of :class:`dict`):
        list of dictionaries of parameters for the regression between :math:`T_i` and
        :math:`T_{i+1}`
      validator (:class:`Validator<TransportMaps.Diagnostic.Validator>`):
        validator to be used to check stability of the solution
      tol (float): target variance diagnostic tolerance
      var_diag_params (dict): parameters to be used in the variance diagnostic approximation
      callback (function): function taking a map and optional additional arguments
        which is called whenever it is deemed necessary by the chosen algorithm
        (e.g. for storing purposes)
      callback_kwargs (dict): additional arguments to be provided to the function
        ``callback``.
      verbosity (int): level of verbosity of the builder
    """
    def __init__(self, base_distribution, target_distribution,
                 transport_map_list, solve_params_list, validator=None,
                 tol=1e-2, var_diag_params=None, laplace_pull=False, 
                 callback=None, callback_kwargs={}, verbosity=0):
        self.solve_counter = 0
        self.transport_map_list = transport_map_list
        self.solve_params_list = solve_params_list
        self.tol = tol
        self.var_diag_params = var_diag_params
        self.laplace_pull = laplace_pull
        super(ToleranceSequentialKullbackLeiblerBuilder,
              self).__init__(
                  base_distribution, target_distribution,
                  transport_map_list[0], solve_params_list[0], validator,
                  callback=callback, callback_kwargs=callback_kwargs,
                  verbosity=verbosity)

    def set_mpi_pool(self, mpi_pool):
        for solve_params in self.solve_params_list:
            solve_params['mpi_pool'] = mpi_pool
        self.var_diag_params['mpi_pool_tuple'] = (None, mpi_pool)
        
    def solve(self, reloading=False):
        r"""
        Returns:
          (:class:`TransportMaps.Maps.TransportMap`) -- the last transport map fitted.
        """
        if not reloading:
            self.solve_counter = 0
        if self.solve_counter == 0:
            if self.var_diag_params is None:
                self.var_diag_params = {
                    'qtype': self.solve_params_list[-1]['qtype'],
                    'qparams': self.solve_params_list[-1]['qparams']}

            self.transport_map = self.transport_map_list[0]
            self.solve_params = self.solve_params_list[0]
                
            if self.laplace_pull:
                # First find Laplace point and center to it
                lap = laplace_approximation(self.target_distribution)
                lap_map = LinearTransportMap.build_from_Gaussian(lap)

                # Set initial conditions to Laplace approximation
                self.transport_map.regression(
                    lap_map, d=base_distribution,
                    qtype=3, qparams=[3]*base_distribution.dim,
                    regularization={'alpha': 1e-4, 'type': 'L2'})

                self.solve_params['x0'] = self.transport_map.coeffs

            tm, log = super(ToleranceSequentialKullbackLeiblerBuilder, self).solve()
            if not log['success']:
                tm.coeffs = x0
                return tm, log
            pull_tar = PullBackTransportMapDistribution(tm, self.target_distribution)
            var = variance_approx_kl(self.base_distribution, pull_tar,
                                     **self.var_diag_params)
            self.logger.info("Variance diagnostic: %e" % var)
            if var <= self.tol:
                return tm, log
            self.solve_counter += 1
        tm_old = self.transport_map_list[self.solve_counter-1]
        for self.transport_map, self.solve_params in zip(
                self.transport_map_list[self.solve_counter:],
                self.solve_params_list[self.solve_counter:]):
            # Here we are assuming nested basis
            for c1, c2 in zip(tm_old.approx_list, self.transport_map.approx_list):
                # Constant part
                for i1, midx1 in enumerate(c1.c.multi_idxs):
                    for i2, midx2 in enumerate(c2.c.multi_idxs):
                        if midx1 == midx2:
                            break
                    c2.c.coeffs[i2] = c1.c.coeffs[i1]
                # Integrated part
                for i1, midx1 in enumerate(c1.h.multi_idxs):
                    for i2, midx2 in enumerate(c2.h.multi_idxs):
                        if midx1 == midx2:
                            break
                    c2.h.coeffs[i2] = c1.h.coeffs[i1]
            
            # solve for the new map using regressed starting point
            self.solve_params['x0'] = self.transport_map.coeffs
            tm, log = super(ToleranceSequentialKullbackLeiblerBuilder, self).solve()
            if not log['success']:
                return tm_old, log
            pull_tar = PullBackTransportMapDistribution(tm, self.target_distribution)
            var = variance_approx_kl(self.base_distribution, pull_tar, **self.var_diag_params)
            self.logger.info("Variance diagnostic: %e" % var)
            if var <= self.tol:
                return tm, log
            tm_old = tm
            self.solve_counter += 1
        # Variance was not met
        log['success'] = False
        log['msg'] = "Desired tolerance was no met by the map adaptivity. " + \
                     "Target variance: %e - Variance: %e " % (self.tol, var)
        return tm, log

class FirstVariationKullbackLeiblerBuilder(KullbackLeiblerBuilder):
    r""" Adaptive builder based on the first variation of the kl divergence

    Given distribution :math:`\nu_\rho` and :math:`\nu_\pi`,
    and the parametric transport map :math:`T[{\bf a}]`,
    provides the functionalities to solve the problem

    .. math::

       \arg\min_{\bf a}\mathcal{D}_{\rm KL}\left(
       T[{\bf a}]_\sharp\rho \Vert \pi\right) =
       \arg\min_{\bf a}\underbrace{\mathbb{E}_\rho\left[
       -\log T[{\bf a}]^\sharp\pi \right]}_{
       \mathcal{J}[T]({\bf x})}

    up to a chosen tolerance, by enriching the map using information
    from the first variation

    .. math::

       \nabla\mathcal{J}[T]({\bf x}) =
       (\nabla_{\bf x}T)^{-\top}
       \left(\log\frac{\rho({\bf x})}{T^\sharp\pi({\bf x})}\right)

    Args:
      base_distribution (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        distribution :math:`\nu_\rho`
      target_distribution (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        distribution :math:`\nu_\pi`
      transport_map (:class:`TransportMap<TransportMaps.Maps.TransportMap>`):
        starting transport map :math:`T`
      validator (:class:`Validator<TransportMaps.Diagnostic.Validator>`):
        validator to be used to check stability of the solution
      eps_bull (float): target tolerance of variance diagnostic
      solve_params (dict): dictionary of parameters for solution
      callback (function): function taking a map and optional additional arguments
        which is called whenever it is deemed necessary by the chosen algorithm
        (e.g. for storing purposes)
      callback_kwargs (dict): additional arguments to be provided to the function
        ``callback``.
      verbosity (int): level of verbosity of the builder
    """
    def __init__(
            self, base_distribution, target_distribution,
            transport_map, validator, eps_bull, regression_builder=None,
            solve_params={}, line_search_params={},
            max_it=20, use_fv_hess=False, 
            prune_trunc=.01, avar_trunc=.1, coeff_trunc=.01,
            callback=None, callback_kwargs={}, verbosity=0):
        self.regression_builder = regression_builder
        self.eps_bull = eps_bull
        self.line_search_params = line_search_params
        self.max_it = max_it
        self.use_fv_hess = use_fv_hess
        self.prune_trunc = prune_trunc
        self.avar_trunc = avar_trunc
        self.coeff_trunc = coeff_trunc
        self.transport_map_list = []
        self.variance_diagnostic_list = []
        super(FirstVariationKullbackLeiblerBuilder,
              self).__init__(
                  base_distribution, target_distribution,
                  transport_map, solve_params, validator,
                  callback=callback, callback_kwargs=callback_kwargs,
                  verbosity=verbosity)
        
    def solve(self, reloading=False):
        if not reloading:
            self.iter_counter = 0
            self.adapt_stage = "validation" # validation, diagnostic, refinement
            self.tolmet = False
        spmet = True
        log = {}
        fv_adapt_status = 'success'
        self.logger.info(
            "Starting. Map structure:\n" + \
            map_structure_str(
                self.transport_map, indent='   ', verbosity=self.verbosity) + \
            map_sparsity_str(
                self.transport_map, indent='   ', verbosity=self.verbosity))
        while self.iter_counter < self.max_it and not self.tolmet:
            self.logger.info("Iteration %d" % self.iter_counter)
            if self.adapt_stage == 'validation':
                spmet = False
                while not spmet and \
                      not log.get('validator_cost_exceeded', False) and \
                      not log.get('validator_fcast_cost_exceeded', False):
                    self.logger.info("Validation...")
                    _, log = super(
                        FirstVariationKullbackLeiblerBuilder,
                        self).solve()
                    if not log.get('success', False) or \
                       log.get('validator_cost_exceeded', False) or \
                       log.get('validator_fcast_cost_exceeded', False):
                        if not log.get('success', True):
                            log['fv_adapt_status'] = 'Failed to converge'
                            self.logger.warning(
                                "KL-minimization failed to converge. " + \
                                "Reverting to the last available map.")
                        else:
                            log['fv_adapt_status'] = 'Cost exceeded'
                            self.logger.warning(
                                "Maximum cost exceeded. Reverting to the last available map.")
                        if len(self.transport_map_list) > 0:
                            return self.transport_map_list[-1], log
                        else:
                            return self.transport_map, log
                    self.transport_map_list.append( self.transport_map )
                    self.callback( self.transport_map, **self.callback_kwargs )
                    spmet = log.get('validator_error', 0.) < \
                            log.get('validator_target_error', np.inf)
                    if not spmet:
                        self.logger.info("Pruning...")
                        # Prune
                        tm_new = prune_map(
                            self.transport_map, log['validator_prune_params'],
                            self.prune_trunc)
                        self.logger.info(
                            "Map pruning. Map structure:\n" + \
                            map_structure_str(tm_new, indent='   ', verbosity=self.verbosity) + \
                            map_sparsity_str(tm_new, indent='   ', verbosity=self.verbosity))
                        if tm_new.n_coeffs == self.transport_map.n_coeffs:
                            # The minimum number of coefficients has already been reached
                            self.logger.warning(
                                "The pruning of the map, did not lead to the removal of " + \
                                "any degree of freedom."
                            )
                            log['fv_adapt_status'] = 'No pruning available'
                            log['fv_adapt_spmet'] = False
                            log['fv_adapt_tolmet'] = False
                            log['fv_adapt_it'] = it
                            if len(self.transport_map_list) > 0:
                                return self.transport_map_list[-1], log
                            else:
                                return self.transport_map, log
                        else:
                            self.transport_map = tm_new
                self.adapt_stage = 'diagnostic'
                self.callback( self.transport_map, **self.callback_kwargs )
            if self.adapt_stage == 'diagnostic':
                # Once the coefficients are determined we check whether the
                # variance diagnostic tolerance is met
                self.logger.info("Computing variance diagnostic...")
                pb_distribution = PullBackTransportMapDistribution(
                    self.transport_map, self.target_distribution)
                var_diag = variance_approx_kl(
                    self.base_distribution, pb_distribution,
                    qtype=self.solve_params['qtype'], qparams=self.solve_params['qparams'],
                    mpi_pool_tuple=(None, self.solve_params.get('mpi_pool')) )
                self.logger.info("Variance diagnostic: %.3e (target %.3e)" % (
                    var_diag,self.eps_bull))
                self.variance_diagnostic_list.append( var_diag )
                if len(self.variance_diagnostic_list) > 1 and \
                   var_diag > self.variance_diagnostic_list[-2]:
                    fv_adapt_status = \
                        "The variance diagnostic is not decreasing. This can be due to " + \
                        "several reasons:\n" + \
                        "   1) the number of quadrature points is insufficient and/or\n" + \
                        "   2) the validation tolerance is too low with respect to the " + \
                        "target adaptivity tolerance and/or\n" + \
                        "   3) a pruning step with too strict tolerance has occurred"
                    self.logger.warning(fv_adapt_status)
                    break
                self.tolmet = var_diag <= self.eps_bull
                self.adapt_stage = 'refinement'
                self.callback( self.transport_map, **self.callback_kwargs )
            if self.adapt_stage == 'refinement':
                if not self.tolmet: # Refinement
                    self.logger.info("Computing first variation...")
                    (x, w) = self.base_distribution.quadrature(
                        self.solve_params['qtype'], self.solve_params['qparams'])
                    # Compute first variation (here we need to make use of the caching)
                    pb_distribution = PullBackTransportMapDistribution(
                        self.transport_map, self.target_distribution)
                    gt, abs_exp_gxgt = compute_first_variation(
                        x, w, self.base_distribution, pb_distribution, self.use_fv_hess,
                        batch_size=self.solve_params.get('batch_size'),
                        mpi_pool=self.solve_params.get('mpi_pool'))
                    # Generate candidate transport map for regression of first variation
                    self.logger.info("Projection of first variation...")
                    fv_tri_tm = first_variation_candidate_triangular_map(
                        self.transport_map, abs_exp_gxgt, self.avar_trunc)
                    fv_tri_tm, log_reg1 = self.regression_builder.solve(
                        fv_tri_tm, gt, x=x, w=w)
                    # TODO: check log success
                    # Line search and evaluation of improved map
                    self.logger.info("Line search...")
                    tm_ev = self.transport_map.evaluate(x)
                    fv_tri_tm_ev = fv_tri_tm.evaluate(x)
                    tm_pxd = self.transport_map.partial_xd(x)
                    fv_tri_tm_pxd = fv_tri_tm.partial_xd(x)
                    delta, ls_success = kl_divergence_fv_line_search(
                        self.target_distribution, w,
                        tm_ev, fv_tri_tm_ev, tm_pxd, fv_tri_tm_pxd,
                        self.line_search_params, mpi_pool=self.solve_params.get('mpi_pool'))
                    if not ls_success:
                        fv_adapt_status = \
                            "Line search did not converge. This may mean several things:\n" + \
                            "   1) the maximum number of line search iterations is too low\n" + \
                            "   2) the validation tolerance is too low to " + \
                            "be able to detect improving directions\n" + \
                            "   3) the validation tolerance is too low to detect that " + \
                            "there is no other improving direction\n" + \
                            "   4) the validation tolerance is too low w.r.t the target " + \
                            "adaptivity tolerance"
                        self.logger.warning(fv_adapt_status)
                        break
                    # Generate candidate transport map for regression on improved map
                    self.logger.info("Generating new candidate map...")
                    tm_new = improved_candidate_map(self.transport_map, fv_tri_tm)
                    x0 = tm_new.coeffs
                    tm_new, log_reg2 = self.regression_builder.solve(
                        tm_new, tm_ev - delta * fv_tri_tm_ev,
                        x=x, w=w, x0=x0)
                    # TODO: check log success
                    # Remove unnecessary coefficients
                    coeffs_weights = np.abs(tm_new.coeffs)
                    coeffs_weights /= max(coeffs_weights)
                    tm_new = prune_map(tm_new, coeffs_weights, self.coeff_trunc)
                    self.logger.info(
                        "Map refinement. Map structure:\n" + \
                        map_structure_str(tm_new, indent='   ', verbosity=self.verbosity) + \
                        map_sparsity_str(tm_new, indent='   ', verbosity=self.verbosity))
                    if is_equal_map(self.transport_map, tm_new):
                        fv_adapt_status = \
                            "The refinement step did not change the original map. " + \
                            "This may be due to several factors:\n" + \
                            "   1) the truncation tolerances are too strict w.r.t. " + \
                            "the validation tolerances\n" + \
                            "   2) the validation tolerance is too low w.r.t the target " + \
                            "adaptivity tolerance"
                        self.logger.warning(fv_adapt_status)
                        break
                    # Set up as the new transport map and new initial conditions
                    self.transport_map = tm_new
                    self.solve_params['x0'] = self.transport_map.coeffs
                self.adapt_stage = 'validation'
            self.iter_counter += 1
            self.callback( self.transport_map, **self.callback_kwargs )
        if self.iter_counter == self.max_it:
            fv_adapt_status = 'Maximum number of iterations exceeded'
        log['fv_adapt_status'] = fv_adapt_status
        log['fv_adapt_tolmet'] = self.tolmet
        log['fv_adapt_spmet'] = spmet
        log['fv_adapt_it'] = self.iter_counter
        return self.transport_map, log

def compute_first_variation(
        x, w, d1, d2, use_fv_hess, batch_size=None, mpi_pool=None):
    # Distribute objects
    d2_distr = dill.loads( dill.dumps(d2) )
    d2_distr.reset_counters()
    mpi_alloc_dmem(d1=d1, d2=d2_distr, mpi_pool=mpi_pool)
    # Link tm to d2.transport_map
    def link_tm_d2(d2):
        return (d2.transport_map,)
    (tm,) = mpi_map_alloc_dmem(
        link_tm_d2, dmem_key_in_list=['d2'], dmem_arg_in_list=['d2'],
        dmem_val_in_list=[d2], dmem_key_out_list=['tm'],
        mpi_pool=mpi_pool)
    # Prepare batch size
    if batch_size is None:
        batch_size = x.shape[0]
    else:
        batch_size = max(
            1 if mpi_pool is None else mpi_pool.nprocs,
            batch_size // d2.dim**2)
    start = 0
    grad_t = np.zeros((x.shape[0], d2.dim))
    abs_exp_gxgt = np.zeros((d2.dim,d2.dim)) if use_fv_hess else None
    while start < x.shape[0]:
        stop = min(x.shape[0], start + batch_size)
        # Compute grad_x and store in distributed memory
        scatter_tuple = (['x'], [x[start:stop,:]])
        (grad_x_tm,) = mpi_map_alloc_dmem(
            'grad_x', scatter_tuple=scatter_tuple, dmem_key_out_list=['grad_x_tm'],
            obj='tm', obj_val=tm, mpi_pool=mpi_pool)
        # Compute first variation and store in memory
        (grad_t_batch,) = mpi_map_alloc_dmem(
            grad_t_kl_divergence, scatter_tuple=scatter_tuple,
            dmem_key_in_list=['d1', 'd2', 'grad_x_tm'],
            dmem_arg_in_list=['d1', 'd2', 'grad_x_tm'],
            dmem_val_in_list=[d1, d2, grad_x_tm],
            dmem_key_out_list=['grad_t'],
            mpi_pool=mpi_pool )
        if mpi_pool is None:
            grad_t[start:stop,:] = grad_t_batch
        else:
            grad_t_list = mpi_pool.get_dmem('grad_t')
            grad_t[start:stop,:] = np.concatenate([gt[0] for gt in grad_t_list], axis=0)
        if use_fv_hess:
            # Compute gradient of first variation
            abs_exp_gxgt += mpi_map(
                grad_x_grad_t_kl_divergence, scatter_tuple=scatter_tuple,
                dmem_key_in_list=['d1', 'd2', 'grad_x_tm', 'grad_t'],
                dmem_arg_in_list=['d1', 'd2', 'grad_x_tm', 'grad_t'],
                dmem_val_in_list=[d1, d2, grad_x_tm, grad_t_batch],
                reduce_obj=AbsExpectationReduce(), reduce_tuple=(['w'], [w[start:stop]]),
                mpi_pool=mpi_pool)
        start = stop
    # Update counters
    if mpi_pool is not None:
        d2_child_list = mpi_pool.get_dmem('d2')
        d2.update_ncalls_tree( d2_child_list[0][0] )
        for (d2_child,) in d2_child_list:
            d2.update_nevals_tree(d2_child)
            d2.update_teval_tree(d2_child)
    # Clear mpi_pool
    if mpi_pool is not None:
        mpi_pool.clear_dmem()
    return grad_t, abs_exp_gxgt
        
def kl_divergence_fv_line_search(
        pi, w, tm_ev, fv_tri_tm_ev, tm_pxd, fv_tri_tm_pxd,
        line_search_params, mpi_pool):

    def objective(x, pi, w, U, pU, mpi_pool):
        # Here x = [T(x)+\delta U(x) | \partial_i T_i(x) + \delta \partial_i U_i(x) ]
        dim = x.shape[1]//2
        TdU = x[:,:dim]
        pTdU = x[:,dim:]
        if np.any(pTdU < 0):
            return np.nan
        lpTdU = np.log(pTdU)
        # Compute pi.log_pdf( T(x)-\delta U(x) )
        scatter_tuple = (['x'], [TdU])
        lpdf = mpi_map('log_pdf', scatter_tuple=scatter_tuple,
                       obj='pi', obj_val=pi, mpi_pool=mpi_pool)
        # Compute log determinant part
        log_det = np.sum(lpTdU, axis=1)
        # Compute output
        out = - np.dot( lpdf + log_det, w )
        return out

    # Distribute pi to the different cores if necessary
    pi_distr = dill.loads( dill.dumps(pi) )
    pi_distr.reset_counters()
    mpi_alloc_dmem(pi=pi, mpi_pool=mpi_pool)

    # Set up line search arguments
    T = tm_ev
    pT = tm_pxd
    U = - fv_tri_tm_ev
    pU = - fv_tri_tm_pxd
    args = (pi, w, U, pU, mpi_pool)
    Tstack = np.hstack( (T, pT) )
    Ustack = np.hstack( (U, pU) )

    # Perform steps of back tracking to find a good delta improving the objective
    # and preserving monotonicity
    maxit = line_search_params.get('maxiter', 20)
    delta = line_search_params.get('delta', 2.)
    isnan = True
    it = 0
    fval0 = objective(Tstack, *args)
    fval = np.inf
    while (isnan or fval > fval0) and it < maxit:
        delta /= 2
        fval = objective(Tstack + delta * Ustack, *args)
        isnan = np.isnan(fval)
        it += 1

    if mpi_pool is not None:
        pi_child_list = mpi_pool.get_dmem('pi')
        pi.update_ncalls_tree( pi_child_list[0][0] )
        for (pi_child,) in pi_child_list:
            pi.update_nevals_tree(pi_child)
            pi.update_teval_tree(pi_child)

    return delta, it < maxit
        
def prune_map(tm, coeffs_weights, coeff_trunc):
    tm = copy.deepcopy(tm)
    # Compute coefficients importance as coeffs * exp_ga_tm
    keep_flags = np.abs(coeffs_weights) > coeff_trunc
    # Run through the components of the map and remove coefficients below threshold
    start = 0
    for d, (avars, comp) in enumerate(zip(tm.active_vars, tm.approx_list)):
        # Constant part
        const_comp = comp.c
        stop = start + const_comp.n_coeffs
        keep_flags_comp = keep_flags[start:stop]
        # Make sure the multi-index (0,0,...,0) is preserved
        zero_idx = const_comp.multi_idxs.index(tuple( [0] * const_comp.dim ))
        keep_flags_comp[zero_idx] = True
        # Remove multi-indices and coefficients
        keep_idxs_comp = list(np.where(keep_flags_comp)[0])
        const_comp.multi_idxs = [ const_comp.multi_idxs[idx] for idx in keep_idxs_comp ]
        const_comp.coeffs = [ const_comp.coeffs[idx] for idx in keep_idxs_comp ]
        start = stop
        # Integrated part
        int_comp = comp.h
        stop = start + int_comp.n_coeffs
        keep_flags_comp = keep_flags[start:stop]
        # Make sure the multi-index (0,0,...,0) is preserved
        zero_idx = int_comp.multi_idxs.index( tuple([0]*int_comp.dim) )
        keep_flags_comp[zero_idx] = True
        # Remove multi-indices and coefficients
        keep_idxs_comp = list(np.where(keep_flags_comp)[0])
        int_comp.multi_idxs = [ int_comp.multi_idxs[idx] for idx in keep_idxs_comp ]
        int_comp.coeffs = [ int_comp.coeffs[idx] for idx in keep_idxs_comp ]
        start = stop
        # Update active variables (union of c and h) and dimensions of c, h, comp
        const_midxs_mat = np.asarray( const_comp.multi_idxs )
        int_midxs_mat = np.asarray( int_comp.multi_idxs )
        keep_idxs = [ i for i in range(comp.dim-1) \
                      if np.any( const_midxs_mat[:,i] != 0 ) \
                      or np.any( int_midxs_mat[:,i] != 0 ) ] + [comp.dim-1]
        const_midxs_mat = const_midxs_mat[:,keep_idxs]
        int_midxs_mat = int_midxs_mat[:,keep_idxs]
        dim = const_midxs_mat.shape[1]
        tm.active_vars[d] = [ avars[idx] for idx in keep_idxs ]
        comp.dim = dim
        const_comp.multi_idxs = mat_idxs_to_list_idxs(const_midxs_mat)
        const_comp.basis_list = [ const_comp.basis_list[idx] for idx in keep_idxs ]
        const_comp.dim = dim
        int_comp.multi_idxs = mat_idxs_to_list_idxs(int_midxs_mat)
        int_comp.basis_list = [ int_comp.basis_list[idx] for idx in keep_idxs ]
        int_comp.dim = dim
    return tm
        
def improved_candidate_map(transport_map, fv_map):
    active_vars = []
    approx_list = []
    for d, (tm_avars, tm_comp,
            fv_avars, fv_comp) in enumerate(zip(
                transport_map.active_vars, transport_map.approx_list,
                fv_map.active_vars, fv_map.approx_list)):
        tm_full_c_blist = tm_comp.c.full_basis_list
        tm_full_h_blist = tm_comp.h.full_basis_list
        # Transform multi-indices into matrices
        tmc_midxs_mat = np.asarray(tm_comp.c.multi_idxs)
        tmh_midxs_mat = np.asarray(tm_comp.h.multi_idxs)
        fv_midxs_mat = np.asarray(fv_comp.multi_idxs)
        # Find active variables to be added to the transport map component
        add_var = [var for var in fv_avars if not var in tm_avars]
        # Add zero columns to the multi-indices of tm_comp
        nmidxs_c = tmc_midxs_mat.shape[0]
        nmidxs_h = tmh_midxs_mat.shape[0]
        for var in add_var:
            ivar = fv_avars.index(var)
            tmc_midxs_mat = np.hstack(
                (tmc_midxs_mat[:,:ivar],
                 np.zeros((nmidxs_c,1), dtype=int),
                 tmc_midxs_mat[:,ivar:]) )
            tmh_midxs_mat = np.hstack(
                (tmh_midxs_mat[:,:ivar],
                 np.zeros((nmidxs_h,1), dtype=int),
                 tmh_midxs_mat[:,ivar:]) )
        # Split fv_comp multi-indices into constant and integrated parts
        c_rows = list(np.where(fv_midxs_mat[:,-1] == 0)[0])
        h_rows = list(set(range(fv_midxs_mat.shape[0])).difference(set(c_rows)))
        fvc_midxs_mat = fv_midxs_mat[c_rows,:]
        fvh_midxs_mat = fv_midxs_mat[h_rows,:]
        # Adjust multi-index for the integrated part
        fvh_midxs_mat[:,-1] -= 1
        # Transform multi-indices into lists
        tmc_midxs = [ tuple(midx) for midx in list(tmc_midxs_mat) ]
        tmh_midxs = [ tuple(midx) for midx in list(tmh_midxs_mat) ]
        fvc_midxs = [ tuple(midx) for midx in list(fvc_midxs_mat) ]
        fvh_midxs = [ tuple(midx) for midx in list(fvh_midxs_mat) ]
        # Enrich transport map component
        add_c_midxs = list( set(fvc_midxs) - set(tmc_midxs) )
        add_h_midxs = list( set(fvh_midxs) - set(tmh_midxs) )
        tmc_midxs += add_c_midxs
        tmh_midxs += add_h_midxs
        # Build basis (using the full basis set provided by the transport map)
        c_basis = [ tm_full_c_blist[a] for a in fv_avars ]
        h_basis = [ tm_full_h_blist[a] for a in fv_avars ]
        # Build constant and integrated linear span functions
        c = LinearSpanApproximation(
            c_basis, spantype='midx', multi_idxs=tmc_midxs,
            full_basis_list=tm_full_c_blist)
        h = LinearSpanApproximation(
            h_basis, spantype='midx', multi_idxs=tmh_midxs,
            full_basis_list=tm_full_h_blist)
        # Set coefficients
        c.coeffs = np.hstack((tm_comp.c.coeffs, np.zeros(len(add_c_midxs))))
        h.coeffs = np.hstack((tm_comp.h.coeffs, np.zeros(len(add_h_midxs))))
        # Assemble component
        comp = type(tm_comp)(c, h)
        # Append to list of components and active variables
        approx_list.append(comp)
        active_vars.append(fv_avars[:])
    new_map = type(transport_map)(
        active_vars, approx_list,
        full_c_basis_list=transport_map.full_c_basis_list,
        full_h_basis_list=transport_map.full_h_basis_list)
    return new_map
        
def first_variation_candidate_triangular_map(
        transport_map, sensitivities, avar_trunc):
    r""" Construct the candidate map to be used in the regression of the first variation.

    It takes the multi-indices in ``transport_map`` and increases them by one,
    adding also active variables if needed.
    The active variables to add are detected using the second moment of the
    first variation, if available.
    """
    active_vars = []
    midxs_list = []
    if sensitivities is not None:
        scaled_sensitivities = sensitivities / np.max(sensitivities)
    else:
        scaled_sensitivities = None
    for d, (tm_comp, tm_comp_avar) in enumerate(zip(
            transport_map.approx_list, transport_map.active_vars)):
        # Update active variables using second order information
        if scaled_sensitivities is not None:
            # tmp = np.abs(exp_gxgt[d,:d+1])
            # scaled_avar_diag = tmp / np.max(tmp)
            # fv_avar = list(np.where(scaled_avar_diag > avar_trunc)[0])
            fv_avar = list(np.where(scaled_sensitivities[d,:d] > avar_trunc)[0])
            add_avar = [ var for var in fv_avar if not var in tm_comp_avar ]
        else:
            add_avar = []
        avar = sorted( tm_comp_avar + add_avar )
        # Adjust multi-index of the exponential part to map to the linear span approx
        h_midxs_mat = np.asarray(tm_comp.h.multi_idxs)
        h_midxs_mat[:,-1] += 1 
        h_midxs = mat_idxs_to_list_idxs(h_midxs_mat)
        # Merge constant and integrated multi-indices
        tm_approx_midxs = tm_comp.c.multi_idxs + h_midxs
        # Transform multi-indices into matrix
        tm_approx_midxs_mat = np.asarray(tm_approx_midxs)
        # Add zero columns to the multi-indices
        for var in add_avar:
            ivar = avar.index(var)
            nmidxs = tm_approx_midxs_mat.shape[0]
            tm_approx_midxs_mat = np.hstack( (tm_approx_midxs_mat[:,:ivar],
                                              np.zeros((nmidxs,1), dtype=int),
                                              tm_approx_midxs_mat[:,ivar:]) )
        # Extract maximum order per direction
        max_ord_vec = np.max(tm_approx_midxs_mat, axis=0)
        max_ord = np.max(np.sum(tm_approx_midxs_mat, axis=1))    
        # Generate new multi-index (GREAT BOTTLENECK)
        max_ord_list = list( max_ord_vec + 1 )
        max_ord += 1
        midxs = generate_total_order_midxs(max_ord_list) # TODO: cython implementation
        # Update
        active_vars.append(avar)
        midxs_list.append(midxs)
    fv_approx = Default_LinearSpanTriangularTransportMap(
        transport_map.dim, midxs_list, active_vars)
    return fv_approx

def mat_idxs_to_list_idxs(mat_idxs):
    return [tuple(idxs) for idxs in list(mat_idxs)]


def map_structure_str(tm, indent, verbosity=0):
    out = indent + "Number of degrees of freedom: %d\n" % tm.n_coeffs
    if verbosity > 0:
        for d, (comp, avars) in enumerate(zip(tm.approx_list, tm.active_vars)):
            out += indent + "Component %3d: Active variables: %s\n" % (d, str(avars))
            out += indent + "   Const part - midxs: %s\n" % (comp.c.multi_idxs)
            if verbosity > 1:
                out += indent + "   Const part - coeffs: %s\n" % (comp.c.coeffs)
            out += indent + "   Integ part - midxs: %s\n" % (comp.h.multi_idxs)
            if verbosity > 1:
                out += indent + "   Integ part - coeffs: %s\n" % (comp.h.coeffs)
    return out

def map_sparsity_str(tm, indent, verbosity=0):
    navar = sum([len(avars) for avars in tm.active_vars])
    totvar = (tm.dim+1)*tm.dim / 2
    out = indent + "Map sparsity: %d/%d (%.4f%%)\n" % (
        navar, totvar, float(navar)/float(totvar)*100)
    if verbosity > 0:
        for d, avars in enumerate(tm.active_vars):
            str_list = [' '] * tm.dim
            for var in avars:
                str_list[var] = 'x'
            out += indent + '   |' + ''.join(str_list) + '|' + '\n'
    return out

def is_equal_map(tm1, tm2):
    match = True
    for d, (c1, a1, c2, a2) in enumerate(zip(
            tm1.approx_list, tm1.active_vars, tm2.approx_list, tm2.active_vars)):
        match = set(a1) == set(a2)
        if not match:
            break
        match = set(c1.c.multi_idxs) == set(c2.c.multi_idxs)
        if not match:
            break
        match = set(c1.h.multi_idxs) == set(c2.h.multi_idxs)
        if not match:
            break
    return match



#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import copy as cp
import numpy as np
import scipy.stats as stats

from TransportMaps.Misc import counted, cached, get_sub_cache
from TransportMaps.Distributions.DistributionBase import *
from TransportMaps.Distributions.FrozenDistributions import GaussianDistribution
from TransportMaps.Distributions.Decomposable.SequentialInferenceDistributions import SequentialHiddenMarkovChainDistribution

from TransportMaps.Functionals.FunctionBase import Function
from TransportMaps.Likelihoods.LikelihoodBase import LogLikelihood

__all__ = [
    'F_phi', 'F_sigma', 'IdentityFunction', 'ConstantFunction',
    'PriorHyperParameters',
    'PriorDynamicsInitialConditions',
    'PriorDynamicsTransition',
    'LogLikelihood',
    'StocVolHyperDistribution',
    'generate_data', 'trim_distribution']

############################
# Transformation functions #
############################

class F_phi(object): 
    def __init__(self,mean,std):
        self.mean = mean
        self.std = std
    def evaluate(self, x):
        Xh = self.mean+self.std*x
        return  2*( np.exp(Xh) ) / (1+np.exp(Xh)) - 1.
    def grad_x(self,x):
        Xh = self.mean+self.std*x
        return  2*self.std*np.exp(Xh) / ( np.exp(Xh) + 1. )**2
    def hess_x(self,x):
        Xh = self.mean+self.std*x
        return  -2*self.std**2*np.exp(Xh) * \
            ( np.exp(Xh) - 1. ) / ( np.exp(Xh) + 1. )**3

class F_sigma(object): #Square root of an inverse gamma
    def __init__(self,k,theta):
        self.k = k
        self.theta = theta
        self.F_Gauss2InvGamma=F_Gauss2InvGamma(k,theta)
    def evaluate(self, x):
        return  np.sqrt( self.F_Gauss2InvGamma.evaluate(x) )
    def grad_x(self,x):
        return  .5*(self.F_Gauss2InvGamma.evaluate(x)**-0.5) * \
            self.F_Gauss2InvGamma.grad_x(x)
    def hess_x(self,x):
        FX=self.F_Gauss2InvGamma.evaluate(x)
        return  -.25*(FX**-1.5)*(self.F_Gauss2InvGamma.grad_x(x)**2) + \
                 .5*(FX**-0.5)*self.F_Gauss2InvGamma.hess_x(x)

class F_Gauss2InvGamma(object):
    def __init__(self,k,theta):
        self.k = k
        self.theta = theta
        self.std = stats.norm()
        self.invGamma = stats.invgamma(k, scale = theta)
    def evaluate(self, X):
        return self.invGamma.ppf( self.std.cdf(X) )
    def grad_x(self,X):
        FX = self.evaluate(X)
        return self.std.pdf(X)/self.invGamma.pdf(FX)
    def hess_x(self,X):
        FX = self.evaluate(X)
        grad_FX = self.grad_x(X)
        a = self.k
        b = self.theta
        gamma = (b - (a+1)*FX )/FX**2
        return - grad_FX*( X + grad_FX*gamma )

class ConstantFunction(Function):
    def __init__(self,constant):
        if not isinstance(constant, float) and not isinstance(constant, int):
            raise ValueError("Constant must be float or int")
        self.constant = float(constant)
        super(ConstantFunction,self).__init__(1)
    def evaluate(self, x):
        return self.constant
    def grad_x(self, x):
        return 0.
    def hess_x(self, x):
        return 0.

class IdentityFunction(Function):
    def __init__(self):
        super(IdentityFunction, self).__init__(1)
    def evaluate(self, x):
        return x
    def grad_x(self, x):
        return 1.
    def hess_x(self, x):
        return 0.


        
################
#    Priors    #
################
        
class PriorHyperParameters(GaussianDistribution):
    r""" Distribution :math:`\pi(\mu, \sigma, \phi)`

    Here :math:`\mu \sim \mathcal{N}(0,\sigma^2_\mu)`,
    :math:`\phi \sim \mathcal{N}(0,1)` and
    :math:`\sigma \sim \mathcal{N}(0,1)`, if they are to be considered
    hyper-parameters.

    Args:
      is_mu_hyper (bool): whether :math:`\mu` is an hyper-parameter
      is_sigma_hyper (bool): whether :math:`\sigma` is an hyper-parameter
      is_phi_hyper (bool): whether :math:`\phi` is an hyper-parameter
      mu_sigma (float): parameter :math:`\sigma_\mu`
    """
    def __init__(self, is_mu_hyper=False, is_sigma_hyper=False, is_phi_hyper=False,
                 sigma_mu=None):
        dim = is_mu_hyper + is_sigma_hyper + is_phi_hyper
        if dim == 0:
            raise ValueError("Avoid defining densities of dimension 0")
        self.is_mu_hyper = is_mu_hyper
        self.is_sigma_hyper = is_sigma_hyper
        self.is_phi_hyper = is_phi_hyper
        self.sigma_mu = sigma_mu
        # Define mean and covariance matrix
        mu = np.zeros(dim)
        cov = np.eye(dim)
        if self.is_mu_hyper:
            cov[0,0] = self.sigma_mu**2.
        # Initialize Gaussian distribution
        super(PriorHyperParameters, self).__init__(mu, sigma=cov)
            
class PriorDynamicsInitialConditions(ConditionalDistribution):
    r""" Conditional distribution :math:`\pi({\bf X}_{t_0}\vert \mu, \sigma, \phi)`

    Args:
      is_mu_hyper (bool): whether :math:`\mu` is an hyper-parameter
      is_sigma_hyper (bool): whether :math:`\sigma` is an hyper-parameter
      is_phi_hyper (bool): whether :math:`\phi` is an hyper-parameter
    """
    def __init__(self,
                 is_mu_hyper=False, mu=None,
                 is_sigma_hyper=False, sigma=None,
                 is_phi_hyper=False, phi=None):
        #Local ordering of the input function: Xmu, Xsigma, Xphi, X0
        self.is_mu_hyper = is_mu_hyper
        self.is_sigma_hyper = is_sigma_hyper
        self.is_phi_hyper = is_phi_hyper
        self.hyper_dim = self.is_mu_hyper + self.is_sigma_hyper + self.is_phi_hyper
        self.state_dim = 1
        super(PriorDynamicsInitialConditions, self).__init__(
            self.state_dim, self.hyper_dim)
        # Init hyper-parameters transformations
        self.mu = mu
        self.phi = phi
        self.sigma = sigma
        counter = 0
        if is_mu_hyper:
            self.index_mu = counter
            counter += 1
        if is_sigma_hyper:
            self.index_sigma = counter
            counter += 1
        if is_phi_hyper:
            self.index_phi = counter
            counter += 1

    def extract_variables(self, x):
        # Extract variables Xmu, Xsigma, Xphi
        (Xmu, Xsigma, Xphi) = (None, None, None)
        if self.is_mu_hyper:
            Xmu = x[:,self.index_mu]
        if self.is_sigma_hyper:
            Xsigma = x[:,self.index_sigma]
        if self.is_phi_hyper:
            Xphi = x[:,self.index_phi]
        return (Xmu, Xsigma, Xphi)

    @cached()
    @counted
    def log_pdf(self, x, y, cache=None, **kwargs):
        X0 = x[:,0]
        Xmu, Xsigma, Xphi = self.extract_variables(y)
        # Evaluate transform functions
        f_Xmu = self.mu.evaluate(Xmu)
        f_Xphi = self.phi.evaluate(Xphi)
        f_Xsigma = self.sigma.evaluate(Xsigma)
        # Evaluate
        return -.5*(1-f_Xphi**2)/f_Xsigma**2*(X0-f_Xmu)**2 + \
            - np.log(f_Xsigma) + 0.5 * np.log(1-f_Xphi**2)

    @cached()
    @counted
    def grad_x_log_pdf(self, x, y, cache=None, **kwargs):
        X0 = x[:,0]
        Xmu, Xsigma, Xphi = self.extract_variables(y)
        # Evaluate transform functions
        f_Xmu = self.mu.evaluate(Xmu)
        f_Xphi = self.phi.evaluate(Xphi)
        f_Xsigma = self.sigma.evaluate(Xsigma)
        # Evaluate gradients of transform functions
        if self.is_phi_hyper:
            grad_f_Xphi = self.phi.grad_x(Xphi)
        if self.is_sigma_hyper:
            grad_f_Xsigma = self.sigma.grad_x(Xsigma)
        # Evaluate
        sdim = self.state_dim
        hdim = self.hyper_dim
        grad = np.zeros( (x.shape[0], sdim + hdim) )
        grad[:,0] = - (1-f_Xphi**2)/f_Xsigma**2*(X0-f_Xmu)
        if self.is_mu_hyper:
            grad[:,sdim+self.index_mu] = (1-f_Xphi**2)/f_Xsigma**2*(X0-f_Xmu)
        if self.is_phi_hyper:
            grad[:,sdim+self.index_phi] = f_Xphi*grad_f_Xphi/f_Xsigma**2*(X0-f_Xmu)**2 + \
                                          - f_Xphi / (1-f_Xphi**2)*grad_f_Xphi
        if self.is_sigma_hyper:
            grad[:,sdim+self.index_sigma] = grad_f_Xsigma / f_Xsigma * \
                                            ( (1-f_Xphi**2)/f_Xsigma**2*(X0-f_Xmu)**2 - 1 )
        return grad

    @counted
    def hess_x_log_pdf(self, x, y, cache=None, **kwargs):
        X0 = x[:,0]
        Xmu, Xsigma, Xphi = self.extract_variables(y)
        # Evaluate transform functions
        f_Xmu = self.mu.evaluate(Xmu)
        f_Xphi = self.phi.evaluate(Xphi)
        f_Xsigma = self.sigma.evaluate(Xsigma)
        # Evaluate gradients and Hessians of transform functions
        if self.is_phi_hyper:
            grad_f_Xphi = self.phi.grad_x(Xphi)
            hess_f_Xphi = self.phi.hess_x(Xphi)
        if self.is_sigma_hyper:
            grad_f_Xsigma = self.sigma.grad_x(Xsigma)
            hess_f_Xsigma = self.sigma.hess_x(Xsigma)
        # Evaluate
        sdim = self.state_dim
        hdim = self.hyper_dim
        hess = np.zeros( (x.shape[0], sdim+hdim, sdim+hdim) )
        hess[:,0,0] = -(1-f_Xphi**2)/f_Xsigma**2
        if self.is_mu_hyper:
            hess[:,sdim+self.index_mu, sdim+self.index_mu] = -(1-f_Xphi**2)/f_Xsigma**2
            hess[:,0,sdim+self.index_mu] = (1-f_Xphi**2)/f_Xsigma**2
            hess[:,sdim+self.index_mu,0] = hess[:,0,sdim+self.index_mu]
            if self.is_sigma_hyper:
                hess[:,sdim+self.index_mu,sdim+self.index_sigma] = \
                    -2*grad_f_Xsigma/f_Xsigma**3*(1-f_Xphi**2)*(X0-f_Xmu)
                hess[:,sdim+self.index_sigma,sdim+self.index_mu] = \
                    hess[:,sdim+self.index_mu,sdim+self.index_sigma]
            if self.is_phi_hyper:
                hess[:,sdim+self.index_mu ,sdim+self.index_phi] = \
                    -2*f_Xphi/f_Xsigma**2*grad_f_Xphi*(X0-f_Xmu)
                hess[:,sdim+self.index_phi,sdim+self.index_mu] = \
                    hess[:,sdim+self.index_mu,sdim+self.index_phi]
        if self.is_sigma_hyper:
            hess[:,sdim+self.index_sigma,sdim+self.index_sigma] = \
                (1-f_Xphi**2)/f_Xsigma**3*(X0-f_Xmu)**2 * \
                (hess_f_Xsigma-3*grad_f_Xsigma**2/f_Xsigma) + \
                grad_f_Xsigma**2/f_Xsigma**2 - hess_f_Xsigma/f_Xsigma
            hess[:,0,sdim+self.index_sigma] = \
                2*grad_f_Xsigma/f_Xsigma**3*(1-f_Xphi**2)*(X0-f_Xmu)
            hess[:,sdim+self.index_sigma,0] = hess[:,0,sdim+self.index_sigma]
            if self.is_phi_hyper:
                hess[:,sdim+self.index_sigma,sdim+self.index_phi]= \
                    -2*f_Xphi/f_Xsigma**3 * grad_f_Xphi * \
                    grad_f_Xsigma*(X0-f_Xmu)**2
                hess[:,sdim+self.index_phi,sdim+self.index_sigma] = \
                    hess[:,sdim+self.index_sigma,sdim+self.index_phi]
        if self.is_phi_hyper:
            hess[:,sdim+self.index_phi,sdim+self.index_phi] = \
                (grad_f_Xphi**2 + f_Xphi*hess_f_Xphi) * \
                ( (X0-f_Xmu)**2/f_Xsigma**2-1./(1.-f_Xphi**2)) - \
                2*f_Xphi**2*grad_f_Xphi**2/(1.-f_Xphi**2)**2
            hess[:,0,sdim+self.index_phi] = 2*f_Xphi/f_Xsigma**2 * \
                                       grad_f_Xphi*(X0-f_Xmu)
            hess[:,sdim+self.index_phi,0] = hess[:,0,sdim+self.index_phi]
        return hess        

class PriorDynamicsTransition(ConditionalDistribution):
    r""" Transition distribution :math:`\pi({\bf X}_{t_{k+1}}\vert {\bf X}_{t_{k}}, \mu, \sigma, \phi)`

    Args:
      is_mu_hyper (bool): whether :math:`\mu` is an hyper-parameter
      is_sigma_hyper (bool): whether :math:`\sigma` is an hyper-parameter
      is_phi_hyper (bool): whether :math:`\phi` is an hyper-parameter
    """
    def __init__(self,
                 is_mu_hyper=False, mu=None,
                 is_sigma_hyper=False, sigma=None,
                 is_phi_hyper=False, phi=None):
        #Local ordering of the input function: Xmu, Xsigma, Xphi, X0
        self.is_mu_hyper = is_mu_hyper
        self.is_phi_hyper = is_phi_hyper
        self.is_sigma_hyper = is_sigma_hyper
        self.hyper_dim = self.is_mu_hyper + self.is_sigma_hyper + self.is_phi_hyper
        self.state_dim = 1
        super(PriorDynamicsTransition, self).__init__(
            self.state_dim, self.state_dim + self.hyper_dim)
        # Init hyper-parameters transformations
        self.mu = mu
        self.phi = phi
        self.sigma = sigma
        counter = 1
        if is_mu_hyper:
            self.index_mu = counter
            counter += 1
        if is_sigma_hyper:
            self.index_sigma = counter
            counter += 1
        if is_phi_hyper:
            self.index_phi = counter
            counter += 1

    def extract_variables(self, x):
        # Extract variables Xt, Xmu, Xsigma, Xphi
        (Xt, Xmu, Xsigma, Xphi) = (x[:,0], None, None, None)
        if self.is_mu_hyper:
            Xmu = x[:,self.index_mu]
        if self.is_phi_hyper:
            Xphi = x[:,self.index_phi]
        if self.is_sigma_hyper:
            Xsigma = x[:,self.index_sigma]
        return (Xt, Xmu, Xsigma, Xphi)

    @cached()
    @counted
    def log_pdf(self, x, y, cache=None, **kwargs):
        Xtp1 = x[:,0]
        (Xt, Xmu, Xsigma, Xphi) = self.extract_variables(y)
        # Evaluate transform functions
        f_Xmu = self.mu.evaluate(Xmu)
        f_Xphi = self.phi.evaluate(Xphi)
        f_Xsigma = self.sigma.evaluate(Xsigma)
        # Evaluate
        return -.5*1/f_Xsigma**2 * \
            ( Xtp1 - f_Xmu - f_Xphi*(Xt - f_Xmu) )**2 - np.log(f_Xsigma)

    @cached()
    @counted
    def grad_x_log_pdf(self, x, y, cache=None, **kwargs):
        Xtp1 = x[:,0]
        (Xt, Xmu, Xsigma, Xphi) = self.extract_variables(y)
        # Evaluate transform functions
        f_Xmu = self.mu.evaluate(Xmu)
        f_Xphi = self.phi.evaluate(Xphi)
        f_Xsigma = self.sigma.evaluate(Xsigma)
        # Evaluate gradients of transform functions
        if self.is_phi_hyper:
            grad_f_Xphi = self.phi.grad_x(Xphi)
        if self.is_sigma_hyper:
            grad_f_Xsigma = self.sigma.grad_x(Xsigma)
        # Evaluate
        sdim = self.state_dim
        hdim = self.hyper_dim
        grad = np.zeros( (x.shape[0], 2*sdim + hdim) )
        grad[:,1] = f_Xphi/f_Xsigma**2*( Xtp1 - f_Xmu - f_Xphi*(Xt - f_Xmu) )
        grad[:,0] = -1/f_Xsigma**2*( Xtp1 - f_Xmu - f_Xphi*(Xt - f_Xmu) )
        if self.is_mu_hyper:
            grad[:,sdim+self.index_mu] = - (f_Xphi-1)/f_Xsigma**2 * \
                                    ( Xtp1 - f_Xmu - f_Xphi*(Xt - f_Xmu) )
        if self.is_sigma_hyper:
            grad[:,sdim+self.index_sigma] = grad_f_Xsigma/f_Xsigma * \
                (1/f_Xsigma**2 * (Xtp1 - f_Xmu - f_Xphi*(Xt - f_Xmu))**2-1)
        if self.is_phi_hyper:
            grad[:,sdim+self.index_phi] =  grad_f_Xphi/f_Xsigma**2 * \
                ( Xt - f_Xmu ) * ( Xtp1 - f_Xmu - f_Xphi*(Xt - f_Xmu) )
        return grad

    @counted
    def hess_x_log_pdf(self, x, y, cache=None, **kwargs):
        Xtp1 = x[:,0]
        (Xt, Xmu, Xsigma, Xphi) = self.extract_variables(y)
        # Evaluate transform functions
        f_Xmu = self.mu.evaluate(Xmu)
        f_Xphi = self.phi.evaluate(Xphi)
        f_Xsigma = self.sigma.evaluate(Xsigma)
        # Evaluate gradients and Hessians of transform functions
        if self.is_phi_hyper:
            grad_f_Xphi = self.phi.grad_x(Xphi)
            hess_f_Xphi = self.phi.hess_x(Xphi)
        if self.is_sigma_hyper:
            grad_f_Xsigma = self.sigma.grad_x(Xsigma)
            hess_f_Xsigma = self.sigma.hess_x(Xsigma)
        # Evaluate
        sdim = self.state_dim
        hdim = self.hyper_dim
        hess = np.zeros( (x.shape[0], 2*sdim+hdim, 2*sdim+hdim) )
        hess[:,1,1] = -(f_Xphi/f_Xsigma)**2
        hess[:,0,0] = -1/f_Xsigma**2
        hess[:,1,0] =  f_Xphi/f_Xsigma**2
        hess[:,0,1] = hess[:,1,0]
        if self.is_mu_hyper:
            hess[:,sdim+self.index_mu,sdim+self.index_mu] =  - ( (f_Xphi-1)/f_Xsigma )**2
            hess[:,1,sdim+self.index_mu] =  (f_Xphi*(f_Xphi-1))/f_Xsigma**2
            hess[:,sdim+self.index_mu,1] = hess[:,1,sdim+self.index_mu]
            hess[:,0,sdim+self.index_mu] =  -(f_Xphi-1)/f_Xsigma**2
            hess[:,sdim+self.index_mu,0] = hess[:,0,sdim+self.index_mu]
            if self.is_sigma_hyper:
                hess[:,sdim+self.index_mu,sdim+self.index_sigma] = \
                    2*(f_Xphi-1)*grad_f_Xsigma/f_Xsigma**3 * \
                    ( Xtp1 - f_Xmu - f_Xphi*(Xt - f_Xmu) )
                hess[:,sdim+self.index_sigma,sdim+self.index_mu] = \
                    hess[:,sdim+self.index_mu,sdim+self.index_sigma]
            if self.is_phi_hyper:
                hess[:,sdim+self.index_mu,sdim+self.index_phi] = \
                    -grad_f_Xphi/f_Xsigma**2 * \
                    ( Xtp1 - f_Xmu - f_Xphi*(Xt - f_Xmu) ) + \
                    (f_Xphi - 1)/f_Xsigma**2*grad_f_Xphi*(Xt - f_Xmu)
                hess[:,sdim+self.index_phi,sdim+self.index_mu] = \
                    hess[:,sdim+self.index_mu,sdim+self.index_phi]
        if self.is_sigma_hyper:
            hess[:,sdim+self.index_sigma,sdim+self.index_sigma] = \
                (hess_f_Xsigma/f_Xsigma - (grad_f_Xsigma/f_Xsigma)**2) * \
                (1/f_Xsigma**2*(Xtp1 - f_Xmu - f_Xphi*(Xt - f_Xmu))**2-1) - \
                2*grad_f_Xsigma**2/f_Xsigma**4*(Xtp1-f_Xmu-f_Xphi*(Xt-f_Xmu))**2
            hess[:,1,sdim+self.index_sigma] = \
                -2*f_Xphi/f_Xsigma**3*grad_f_Xsigma * \
                ( Xtp1 - f_Xmu - f_Xphi*(Xt - f_Xmu) )
            hess[:,sdim+self.index_sigma,1] = hess[:,1,sdim+self.index_sigma]
            hess[:,0,sdim+self.index_sigma] = \
                2*1/f_Xsigma**3*grad_f_Xsigma*( Xtp1 - f_Xmu - f_Xphi*(Xt - f_Xmu) )
            hess[:,sdim+self.index_sigma,0] = hess[:,0,sdim+self.index_sigma]
            if self.is_phi_hyper:
                hess[:,sdim+self.index_sigma,sdim+self.index_phi]= \
                    -2*grad_f_Xsigma*grad_f_Xphi/f_Xsigma**3*(Xt-f_Xmu) * \
                    ( Xtp1 - f_Xmu - f_Xphi*(Xt - f_Xmu) )
                hess[:,sdim+self.index_phi,sdim+self.index_sigma] = \
                    hess[:,sdim+self.index_sigma,sdim+self.index_phi]
        if self.is_phi_hyper:
            hess[:,sdim+self.index_phi,sdim+self.index_phi] = \
                hess_f_Xphi/f_Xsigma**2*(Xt-f_Xmu) * \
                ( Xtp1 - f_Xmu - f_Xphi*(Xt - f_Xmu) ) - \
                (grad_f_Xphi/f_Xsigma)**2*(Xt - f_Xmu)**2
            hess[:,1,sdim+self.index_phi] = \
                grad_f_Xphi/f_Xsigma**2*( Xtp1 - f_Xmu - f_Xphi*(Xt - f_Xmu) ) \
                - f_Xphi * grad_f_Xphi / f_Xsigma**2*(Xt - f_Xmu)
            hess[:,sdim+self.index_phi,1] = hess[:,1,sdim+self.index_phi]
            hess[:,0,sdim+self.index_phi] = grad_f_Xphi/f_Xsigma**2*(Xt - f_Xmu)
            hess[:,sdim+self.index_phi,0] = hess[:,0,sdim+self.index_phi]
        return hess


##################
#   Likelihood   #        
##################
        
class LogLikelihood(LogLikelihood):
    r""" Abstract class for log-likelihood :math:`\log \pi({\bf y}_t \vert {\bf X}_t), \mu, \phi, \sigma`

    where, up to integration constants,

    .. math::

       \log\pi({\bf y}_t \vert {\bf X}_t, \mu, \phi, \sigma) = - \frac{1}{2}({\bf y}_t^2 \exp(-{\bf X}_t))
    
    Args:
      y (:class:`ndarray<numpy.ndarray>`): data
      is_mu_hyper (bool): whether :math:`\mu` is an hyper-parameter
      is_phi_hyper (bool): whether :math:`\phi` is an hyper-parameter
      is_sigma_hyper (bool): whether :math:`\sigma` is an hyper-parameter
    """
    def __init__(self, y,
                 is_mu_hyper=False, is_sigma_hyper=False, is_phi_hyper=False):
        #Local ordering of the input function: X0, Xmu, Xsigma, Xphi
        self.is_mu_hyper = is_mu_hyper
        self.is_sigma_hyper = is_phi_hyper
        self.is_phi_hyper = is_sigma_hyper
        self.hyper_dim = self.is_mu_hyper + self.is_sigma_hyper + self.is_phi_hyper
        self.state_dim = 1
        dim = self.state_dim + self.hyper_dim
        super(LogLikelihood, self).__init__(y, dim)

    def extract_variables(self, x):
        # Extract variable Xt out of (Xt, Xmu, Xsigma, Xphi)
        Xt = x[:,0]
        return Xt

    @cached()
    @counted
    def evaluate(self, x, cache=None, **kwargs):
        Xt = self.extract_variables(x)
        return -.5 * self.y**2. * np.exp(-Xt) -.5 * Xt

    @cached()
    @counted
    def grad_x(self, x, cache=None, **kwargs):
        Xt = self.extract_variables(x)
        out = np.zeros((x.shape[0], self.dim))
        out[:,0] = .5 * self.y**2. * np.exp(-Xt) - .5
        return out

    @counted
    def hess_x(self, x, cache=None, **kwargs):
        Xt = self.extract_variables(x)
        out = np.zeros((x.shape[0], self.dim, self.dim))
        out[:,0,0] = - .5 * self.y**2. * np.exp(-Xt)
        return out


##################
# Full Posterior #
##################

class StocVolHyperDistribution(SequentialHiddenMarkovChainDistribution):
    def __init__(self,
                 is_mu_hyper=False, is_sigma_hyper=False, is_phi_hyper=False,
                 mu=None, sigma=None, phi=None, 
                 mu_sigma=1.,
                 sigma_k=1., sigma_theta=.1,
                 phi_mean=3., phi_std=1.):
        # Initialize true dynamic (optionally provided)
        self.Xt = None
        # Build prior on hyper parameters if necessary
        self.is_mu_hyper = is_mu_hyper
        self.is_phi_hyper = is_phi_hyper
        self.is_sigma_hyper = is_sigma_hyper
        hyper_dim = is_mu_hyper + is_phi_hyper + is_sigma_hyper
        if hyper_dim > 0:
            pi_hyper = PriorHyperParameters(is_mu_hyper, is_sigma_hyper, is_phi_hyper,
                                            mu_sigma)
        else:
            pi_hyper = None
        # Store transformations
        if is_mu_hyper:
            self.mu = IdentityFunction()
        else:
            self.mu = ConstantFunction(mu)
        if is_phi_hyper:
            self.phi = F_phi(phi_mean, phi_std)
        else:
            self.phi = ConstantFunction(phi)
        if is_sigma_hyper:
            self.sigma = F_sigma(sigma_k, sigma_theta)
        else:
            self.sigma = ConstantFunction(sigma)
        # Initialize distribution (no transitions defined yet)
        super(StocVolHyperDistribution, self).__init__([], [], pi_hyper)

    @counted
    def action_hess_x_log_pdf(self, x, dx, cache=None, **kwargs):
        H = self.hess_x_log_pdf(x, cache=cache)
        return np.einsum('...ij,...j->...i', H, dx)

    @property
    def index_mu(self):
        if self.is_mu_hyper:
            return 0
        else:
            return None
    @property
    def index_sigma(self):
        if self.is_sigma_hyper:
            counter = len([o for o in [self.index_mu] if o is not None])
            return counter
        else:
             return None
    @property
    def index_phi(self):
        if self.is_phi_hyper:
            counter = len([o for o in [self.index_mu, self.index_sigma]
                           if o is not None])
            return counter
        else:
            return None

    def get_tvec(self):
        return np.arange(self.get_nsteps())

    def get_nsteps(self):
        return len(self.pi_list)

    def assimilate(self, y=None, Xt=None):
        r""" Assimilate one piece of data.

        Args:
          y (:class:`ndarray<numpy.ndarray>`): data. ``y==None`` stands for missing data.
          Xt (:class:`ndarray<numpy.ndarray>`): true dynamics
        """
        ll = None
        if y is not None:
            ll = LogLikelihood(y, self.is_mu_hyper,
                               self.is_sigma_hyper, self.is_phi_hyper)
        if len(self.pi_list) == 0: # Initial conditions
            pi = PriorDynamicsInitialConditions(
                self.is_mu_hyper, self.mu,
                self.is_sigma_hyper, self.sigma,
                self.is_phi_hyper, self.phi)
            if Xt is not None:
                self.Xt = [Xt]
        else: # Transitions
            pi = PriorDynamicsTransition(
                self.is_mu_hyper, self.mu,
                self.is_sigma_hyper, self.sigma,
                self.is_phi_hyper, self.phi)
            if Xt is not None:
                if self.Xt is None or len(self.Xt) != len(self.pi_list):
                    raise ValueError("The true dynamics must be provided " + \
                                     "at all steps, if one wish to privide them")
                self.Xt.append(Xt)
        self.append(pi, ll)

def generate_data(nsteps, mu = None, sigma = None, phi = None):
    #Random initial conditions
    x0 = sigma/np.sqrt(1.-phi**2)*np.random.randn(1) + mu
    #Noise observations
    noise_obs = np.random.randn(nsteps)
    #Noise dynamics
    noise_dyn = np.random.randn(nsteps)
    # Allocate variables
    Xt = np.zeros(nsteps)
    Xt[0] = x0
    # Simulate
    for ii in range(nsteps-1):
        Xt[ii+1] = mu+phi*(Xt[ii]-mu) + sigma*noise_dyn[ii]
    #Collect observations
    dataObs = list(noise_obs*np.exp(.5*Xt))
    return (dataObs, Xt)

def trim_distribution(old_dens, nsteps):
    old_nsteps = old_dens.get_nsteps()
    if nsteps > old_nsteps:
        raise ValueError("The number of steps can only be decreased with this function")
    new_dens = cp.deepcopy(old_dens)
    # re-init distribution
    new_dens.dim = new_dens.hyper_dim
    new_pi_list = new_dens.pi_list[:nsteps]
    new_ll_list = new_dens.ll_list[:nsteps]
    new_dens.pi_list = []
    new_dens.ll_list = []
    for pi, ll in zip(new_pi_list, new_ll_list):
        new_dens.append(pi, ll)
    return new_dens
    
    



#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import unittest
import numpy as np
import numpy.random as npr
import numpy.linalg as npla

import TransportMaps as TM
import TransportMaps.Diagnostics as DIAG
import TransportMaps.Likelihoods as LKL
import TransportMaps.Distributions as DIST
import TransportMaps.Distributions.Inference as INFDIST

try:
    import mpi_map
    MPI_SUPPORT = True
except:
    MPI_SUPPORT = False

class LaplaceApproximation(unittest.TestCase):

    def setUp(self):
        npr.seed(1)
        self.tol = 1e-8

    def compute_laplace(self, d):
        import numpy as np
        import numpy.linalg as npla
        import TransportMaps as TM

        # Build Laplace approximation
        d_laplace = TM.laplace_approximation(d, None, tol=self.tol)

        # Compare mean
        self.assertTrue( npla.norm(d.mu - d_laplace.mu) <= self.tol )

        # Compute KL divergence between the two
        qtype = 3
        qparams = [10]*d.dim
        kldiv = TM.kl_divergence(d, d_laplace, None, None,
                                 qtype=qtype, qparams=qparams)
        self.assertTrue( np.abs(kldiv) <= 1e-2 )

    def test_linear1d(self):
        import TransportMaps.tests.TestFunctions as TF
        title, setup, Tparams = TF.get(0)
        self.compute_laplace(Tparams['target_distribution'])

    def test_linear2d(self):
        import TransportMaps.tests.TestFunctions as TF
        title, setup, Tparams = TF.get(9)
        self.compute_laplace(Tparams['target_distribution'])

    def test_gauss_low_rank_nd(self):
        #################
        # Construct distribution
        dx = 40
        dy = 10
        # Prior
        mux = npr.randn(dx)
        sigmax = npr.randn(dx**2).reshape((dx,dx))
        sigmax = np.dot(sigmax, sigmax.T)
        prior = DIST.GaussianDistribution(mux, sigmax)
        # Likelihood
        mu = npr.randn(dy)
        sigma = npr.randn(dy**2).reshape((dy,dy))
        sigma = np.dot(sigma, sigma.T)
        c = npr.randn(dy)
        T = npr.randn(dy*dx).reshape((dy,dx))
        y = c + np.dot(T, prior.rvs(1)[0,:])
        logL = LKL.AdditiveLinearGaussianLogLikelihood(y, c, T, mu, sigma)
        # Posterior
        pi = INFDIST.BayesPosteriorDistribution(logL, prior)

        ##################
        # Build Laplace approximation using action of Hessian
        pi_laplace = TM.laplace_approximation(pi, None, tol=self.tol, ders=2,
                                              hessact=True)
        # Check variance diagnostic (should be at machine precision)
        var_diag = DIAG.variance_approx_kl(pi_laplace, pi, qtype=0, qparams=10000)
        self.assertTrue( var_diag < 1e-13 )

def build_suite(ttype='all'):
    suite_laplace_approx = unittest.TestLoader().loadTestsFromTestCase( LaplaceApproximation )
    # GROUP SUITES
    suites_list = []
    if ttype in ['all','serial']:
        suites_list = [ suite_laplace_approx ]
    all_suites = unittest.TestSuite( suites_list )
    return all_suites


def run_tests(ttype='all'):
    all_suites = build_suite(ttype)
    # RUN
    unittest.TextTestRunner(verbosity=2).run(all_suites)

if __name__ == '__main__':
    run_tests()



#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

__all__ = [
    'check_der', 'check_grad_a', 'check_hess_a_from_grad_a',
    'check_action_hess_a_from_grad_a', 'check_grad_x',
    'check_hess_x_from_grad_x', 'grad_a_fd', 'grad_x_fd',
    'grad_x_fx',]

import numpy as np
import time

def check_der(f, dx_f, x, dx, params={}, title='', verbose=True):
    maxerr = np.max(np.abs(dx_fdapprox(f,x,dx,params) - dx_f(x,params)))
    if verbose:
        print("Check der %s - Max err: %e" % (title,maxerr))

def check_grad_a(f, grad_f, x, dx, params={}, end=True, title='', verbose=True):
    app_start = time.clock()
    app = grad_a_fd(f,x,dx,params,end)
    app_stop = time.clock()
    app_time = app_stop-app_start
    exa_start = time.clock()
    exa = grad_f(x, **params)
    exa_stop = time.clock()
    exa_time = exa_stop - exa_start
    err = np.abs(app-exa).flatten()
    idxmer = np.argmax(err)
    maxerr = err[idxmer]
    if not np.allclose(exa, app, rtol=100*dx, atol=10*dx):
        print('Max err: %e. Abs: %e. rtol: %e. atol: %e' % (
            maxerr, exa.flatten()[idxmer], 100*dx, 10*dx))
    if verbose:
        print("Check grad %s - Max err: %e - FD time: %.4f - Analytic time: %.4f" % (
            title,maxerr,app_time,exa_time))
    return np.allclose(exa, app, rtol=100*dx, atol=10*dx)

def check_hess_a_from_grad_a(grad_f, hess_f, x, dx, params={}, title='', verbose=True):
    exa_start = time.clock()
    exa = hess_f(x, **params)
    exa_stop = time.clock()
    exa_time = exa_stop - exa_start
    app_start = time.clock()
    app = grad_a_fd(grad_f, x, dx, params)
    app_stop = time.clock()
    app_time = app_stop-app_start
    err = np.abs(app-exa).flatten()
    maxerr = np.max(err)
    if verbose:
        print("Check hess %s - Max err: %e - FD time: %.4f - Analytic time: %.4f" % (
            title,maxerr, app_time,exa_time))
    return np.allclose(exa, app, rtol=100*dx, atol=10*dx)

def check_action_hess_a_from_grad_a(grad_f, action_hess_f, x, dx, v, params={}, title='', verbose=True):
    exa_start = time.clock()
    exa = action_hess_f(x, v, **params)
    exa_stop = time.clock()
    exa_time = exa_stop - exa_start
    app_start = time.clock()
    app = grad_a_fd(grad_f, x, dx, params)
    app = np.dot(app, v)
    app_stop = time.clock()
    app_time = app_stop-app_start
    err = np.abs(app-exa).flatten()
    maxerr = np.max(err)
    if verbose:
        print("Check hess %s - Max err: %e - FD time: %.4f - Analytic time: %.4f" % (
            title,maxerr,app_time,exa_time))
    return np.allclose(exa, app, rtol=100*dx, atol=10*dx)
    
def check_grad_x(f, grad_f, x, dx, params={}, title='', verbose=True):
    app_start = time.clock()
    app = grad_x_fd(f,x,dx,params)
    app_stop = time.clock()
    app_time = app_stop-app_start
    exa_start = time.clock()
    exa = grad_f(x, **params)
    exa_stop = time.clock()
    exa_time = exa_stop - exa_start
    err = np.abs(app-exa).flatten()
    maxerr = np.max(err)
    if verbose:
        print("Check grad %s - Max err: %e - FD time: %.4f - Analytic time: %.4f" % (
            title,maxerr,app_time,exa_time))
    return np.allclose(exa, app, rtol=dx, atol=dx)

def check_hess_x_from_grad_x(grad_f, hess_f, x, dx, params={}, title='', verbose=True):
    exa_start = time.clock()
    exa = hess_f(x, **params)
    exa_stop = time.clock()
    exa_time = exa_stop - exa_start
    app_start = time.clock()
    app = grad_x_fd(grad_f, x, dx, params)
    app_stop = time.clock()
    app_time = app_stop-app_start
    err = np.abs(app-exa).flatten()
    maxerr = np.max(err)
    if verbose:
        print("Check hess %s - Max err: %e - FD time: %.4f - Analytic time: %.4f" % (title,maxerr,
                                                                                     app_time,exa_time))
    # return np.allclose(exa, app, rtol=100*dx, atol=10*dx)
    return np.allclose(exa, app, rtol=dx, atol=dx)

def grad_a_fd(f, x, dx, params={}, end=True):
    tmp = f(x, **params)
    if isinstance(tmp, float): extradims = 0
    else: extradims = tmp.ndim
    if end:
        out = np.zeros(tmp.shape + (len(x),))
    else:
        out = np.zeros(tmp.shape[:1] + (len(x),) + tmp.shape[1:])
    idxbase = tuple( [slice(None)]* extradims )
    for i in range(len(x)):
        xc_minus = x.copy()
        xc_plus = x.copy()
        xc_minus[i] -= dx/2.
        xc_plus[i] += dx/2.
        fm = f(xc_minus, **params)
        fp = f(xc_plus, **params)
        if end:
            out[idxbase + (i,)] = (fp-fm)/dx
        else:
            idx = (idxbase[:1] + (i,) + idxbase[1:])
            out[idx] = (fp-fm)/dx
    return out

def grad_x_fd(f, x, dx, params={}):
    r""" Compute :math:`\nabla_{\bf x} f({\bf x})` of :math:`f:\mathbb{R}^d\rightarrow\mathbb{R}`
    """
    tmp = f(x, **params)
    if isinstance(tmp, float): extradims = 0
    else: extradims = tmp.ndim
    nsamp = x.shape[0]
    dim = x.shape[1]
    out = np.zeros(tmp.shape + (dim,))
    idxbase = tuple( [slice(None)]* extradims )
    for i in range(dim):
        xcm = x.copy()
        xcp = x.copy()
        xcm[:,i] -= dx/2.
        xcp[:,i] += dx/2.
        fm = f(xcm, **params)
        fp = f(xcp, **params)
        out[idxbase + (i,)] = (fp-fm)/dx
    return out

def grad_x_fx(f, x, dx, params={}):
    r""" Compute :math:`\nabla_{\bf x} f_{\bf x}({\bf x})` of :math:`f_{\bf x}:\mathbb{R}^d\rightarrow\mathbb{R}^d`
    """
    nsamp = x.shape[0]
    dim = x.shape[1]
    out = np.zeros((nsamp,dim,dim))
    for i in range(dim):
        xcm = x.copy()
        xcp = x.copy()
        xcm[:,i] -= dx/2.
        xcp[:,i] += dx/2.
        fm = f(xcm, **params)
        fp = f(xcp, **params)
        out[:,:,i] = (fp-fm)/dx
    return out



#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import numpy as np
import numpy.linalg as npla
import numpy.random as npr
import scipy.stats as stats
import scipy.optimize as sciopt
import scipy.linalg as scila

from TransportMaps import mpi_map, distributed_sampling
from TransportMaps.ObjectBase import TMO
from TransportMaps.Distributions.TransportMapDistributions import \
    PullBackTransportMapDistribution

__all__ = ['KLMinimizationValidator',
           'SampleAverageApproximationKLMinimizationValidator',
           'GradientChi2KLMinimizationValidator',
           'GradientStabilityKLMinimizationValidator']

class KLMinimizationValidator(TMO):
    def __init__(
            self, eps_sp, cost_function, max_cost, eps_sp_rel=0,
            max_nsamps=np.inf, stop_on_fcast=False):
        super(KLMinimizationValidator, self).__init__()
        self.eps_sp_abs = eps_sp
        self.eps_sp_rel = eps_sp_rel
        self.cost_function = cost_function
        self.max_cost = max_cost
        self.max_nsamps = max_nsamps
        self.stop_on_fcast = stop_on_fcast

    def solve_to_tolerance(
            self, transport_map, base_distribution, target_distribution,
            solve_params):
        log = {}
        max_nsamps_flag = False
        cost = self.cost_function(
            ncalls=getattr(target_distribution, 'ncalls', {}),
            nevals=getattr(target_distribution, 'nevals', {}),
            teval=getattr(target_distribution, 'teval', {}))
        cost_flag = cost > self.max_cost
        fcast_cost = 0.
        fcast_cost_flag = False
        err = np.inf
        err_mag = 0
        target_err = err_mag * self.eps_sp_rel + self.eps_sp_abs
        prune_params = None

        pull_tar = PullBackTransportMapDistribution(
            transport_map, target_distribution)
        ncalls_x_solve = getattr(target_distribution, 'ncalls', {}).copy()
        lst_ncalls = getattr(target_distribution, 'ncalls', {}).copy()

        it = 1
        while err > target_err and not cost_flag:
            if it > 1:
                # Compute refinement and forecast cost flag
                solve_params['qparams'], nsamps = self.refinement(
                    base_distribution, pull_tar,
                    qtype=solve_params['qtype'], qparams=solve_params['qparams'],
                    err_mag=err_mag, ref_params=ref_params)
                self.logger.info("Refinement - nsamps: %d" % nsamps)
                if nsamps >= self.max_nsamps:
                    max_nsamps_flag = True
                    self.logger.warning("Maximum number of samples reached.")
                    break
                if self.stop_on_fcast:
                    fcast_cost = self.cost_function(
                        ncalls=target_distribution.ncalls,
                        nevals=target_distribution.nevals,
                        teval=target_distribution.teval,
                        ncalls_x_solve=ncalls_x_solve, new_nx=nsamps)
                    fcast_cost_flag = fcast_cost > self.max_cost
                    if fcast_cost_flag: # Stop
                        self.logger.warning("Predicted cost exceeds maximum cost allowed.")
                        break
                solve_params['x0'] = transport_map.coeffs # Warm start
            # Solve
            log = transport_map.minimize_kl_divergence(
                base_distribution, pull_tar, **solve_params)
            if not log['success']:
                cost = self.cost_function(
                    ncalls=target_distribution.ncalls,
                    nevals=target_distribution.nevals,
                    teval=target_distribution.teval)
                break
            # Estimate error and compute cost flag
            solve_params['x0'] = transport_map.coeffs # Warm start
            err, err_mag, ref_params, prune_params = self.error_estimation(
                base_distribution, pull_tar, solve_params, full_output=True)
            target_err = err_mag * self.eps_sp_rel + self.eps_sp_abs
            ncalls_x_solve = {
                key: value - lst_ncalls.get(key,0)
                for key, value in pull_tar.ncalls.items()}
            lst_ncalls = pull_tar.ncalls.copy()
            cost = self.cost_function(
                ncalls=target_distribution.ncalls,
                nevals=target_distribution.nevals,
                teval=target_distribution.teval)
            cost_flag = cost > self.max_cost
            self.print_info(err, target_err, cost, solve_params, ref_params, prune_params)
            it += 1

        log['validator_max_nsamps_exceeded'] = max_nsamps_flag
        log['validator_cost'] = cost
        log['validator_cost_exceeded'] = cost_flag
        log['validator_fcast_cost'] = fcast_cost
        log['validator_fcast_cost_exceeded'] = fcast_cost_flag
        log['validator_error'] = err
        log['validator_prune_params'] = prune_params
        log['validator_target_error'] = target_err
        return log

    def print_info(self, err, target_err, cost, solve_params, ref_params, prune_params):
        self.logger.info(
            "nsamps: %d - " % solve_params['qparams'] + \
            "err: %.3e (target: %.3e)" % (err, target_err) + \
            "- cost: %.2e" % cost)
        
    def error_estimation(self, base_distribution, pull_tar, solve_params):
        raise NotImplementedError("To be implemented in subclasses")

    def refinement(self, base_distribution, pull_tar,
                   qtype, qparams, ref_params):
        r"""
        Returns:
          (:class:`tuple`) -- containing the new ``qparams`` and the
            number of points corresponding to it.
        """
        raise NotImplementedError("To be implemented in subclasses")

class SampleAverageApproximationKLMinimizationValidator(KLMinimizationValidator):
    def __init__(self, eps_sp_rel, eps_sp_abs, cost_function, max_cost,
                 max_nsamps=np.inf, stop_on_fcast=False,
                 upper_mult=10, lower_n=2, alpha=0.05,
                 lmb_def=2, lmb_max=10):
        if upper_mult < 1:
            raise AttributeError("The upper_mult argument must be a float >= 1")
        if lower_n < 2:
            raise AttributeError("The lower_n argument must be an integer >= 2")
        self.upper_mult = upper_mult
        self.lower_n = lower_n
        self.alpha = alpha
        self.lmb_def = lmb_def
        self.lmb_max = lmb_max
        super(SampleAverageApproximationKLMinimizationValidator,
              self).__init__(
                  eps_sp_abs, cost_function, max_cost, eps_sp_rel=eps_sp_rel,
                  max_nsamps=max_nsamps, stop_on_fcast=stop_on_fcast)

    def print_info(self, err, target_err, cost, solve_params, ref_params, prune_params):
        self.logger.info(
            "nsamps: %d " % solve_params['qparams'] + \
            "- err: %.3e [L: %.3e, U:%.3e] (target: %.3e) " % (
                err, ref_params['lower_interval'],
                ref_params['upper_interval'], target_err) + \
            "- cost: %.2e" % cost)
        
    def error_estimation(
            self, base_distribution, pull_tar, solve_params, full_output=False):
        if solve_params['qtype'] != 0:
            raise AttributeError(
                "The Sample Average Approximation validator is defined only for " + \
                "Monte Carlo quadrature rules")
        tm = pull_tar.transport_map
        # Compute upper bound (distributed sampling)
        upper_nsamps = int(np.ceil(self.upper_mult * solve_params['qparams']))
        (x, w) = distributed_sampling(
            base_distribution, 0, upper_nsamps, mpi_pool=solve_params.get('mpi_pool'))
        v2 = - mpi_map(
            "log_pdf", obj=pull_tar,
            dmem_key_in_list=['x'],
            dmem_arg_in_list=['x'],
            dmem_val_in_list=[x],
            mpi_pool=solve_params.get('mpi_pool'))
        if solve_params['regularization'] is not None:
            if solve_params['regularization']['type'] == 'L2':
                v2 += solve_params['regularization']['alpha'] * \
                      npla.norm(tm.coeffs - tm.get_identity_coeffs(), 2)**2.
        upper_mean = np.sum( v2 ) / float(upper_nsamps)
        upper_var = np.sum( (v2 - upper_mean)**2 ) / float(upper_nsamps-1)
        upper_interval = stats.t.ppf(1-self.alpha, upper_nsamps) * \
                         np.sqrt(upper_var/float(upper_nsamps))
        upper_bound = upper_mean + upper_interval
        # Compute lower bound
        res_coeffs = pull_tar.coeffs[:]
        lower_nsamps = self.lower_n
        v2 = np.zeros(lower_nsamps)
        coeffs = np.zeros((lower_nsamps, tm.n_coeffs))
        for i in range(lower_nsamps):
            log = tm.minimize_kl_divergence(
                base_distribution, pull_tar, **solve_params)
            v2[i] = log['fval']
            coeffs[i,:] = tm.coeffs
        lower_mean = np.sum(v2) / float(lower_nsamps)
        lower_var = np.sum((v2 - lower_mean)**2) / float(lower_nsamps-1)
        lower_interval = stats.t.ppf(1-self.alpha, lower_nsamps) * \
                         np.sqrt(lower_var/float(lower_nsamps))
        lower_bound = lower_mean - lower_interval
        pull_tar.coeffs = res_coeffs # Restore coefficients
        # Error as the gap between the bounds
        err = max(0., upper_bound - lower_bound)
        if full_output:
            # Refinement parameters
            ref_params = {
                'upper_nsamps': upper_nsamps,
                'upper_mean': upper_mean,
                'upper_var': upper_var,
                'upper_interval': upper_interval,
                'lower_nsamps': lower_nsamps,
                'lower_mean': lower_mean,
                'lower_var': lower_var,
                'lower_interval': lower_interval
            }
            # Pruning parameter as 1/stand.dev. of coefficients scaled to (0,1)
            std_coeffs = np.std(coeffs, axis=0)
            prune_params = (1/std_coeffs)/max(1/std_coeffs)
            return err, upper_mean, ref_params, prune_params
        else:
            return err

    def refinement(self, base_distribution, pull_tar,
                   qtype, qparams, err_mag, ref_params):
        if qtype != 0:
            raise AttributeError(
                "The Sample Average Approximation validator is defined only for " + \
                "Monte Carlo quadrature rules")
        X = (self.eps_sp_abs + self.eps_sp_rel * err_mag) - \
            (ref_params['upper_mean'] - ref_params['lower_mean']) - \
            stats.t.ppf(1-self.alpha, ref_params['upper_nsamps']) * \
            np.sqrt(ref_params['upper_var']/ref_params['upper_nsamps'])
        if X > 0:
            def f(m, alpha, X, lv):
                return stats.t.ppf(1-alpha, m) / m * np.sqrt(lv) - X
            if np.sign(
                    f(ref_params['lower_nsamps'], self.alpha,
                      X, ref_params['lower_var']) ) * \
                np.sign(
                    f(self.lmb_max, self.alpha,
                      X, ref_params['lower_var']) ) == -1:
                m = sciopt.bisect(
                    f, ref_params['lower_nsamps'], self.lmb_max,
                    args=(self.alpha, X, ref_params['lower_var']))
                q = int(np.ceil( m/ref_params['lower_nsamps'] * qparams ))
            else:
                q = int( np.ceil(self.lmb_def * qparams) )
        else:
            q = int( np.ceil(self.lmb_def * qparams) )
        q = min(q, self.max_nsamps)
        return q, q

class GradientChi2KLMinimizationValidator(KLMinimizationValidator):
    def __init__(self, eps_sp, cost_function, max_cost,
                 max_nsamps=np.inf, stop_on_fcast=False,
                 n_grad_samps=10, n_bootstrap=None,
                 alpha=0.95, lmb_def=2, lmb_max=10,
                 fungrad=False):
        self.n_grad_samps = n_grad_samps
        self.n_bootstrap = n_bootstrap
        self.alpha = alpha
        self.lmb_def = lmb_def
        self.lmb_max = lmb_max
        self.fungrad = fungrad
        super(GradientChi2KLMinimizationValidator,
              self).__init__(eps_sp, cost_function, max_cost, max_nsamps, stop_on_fcast)
        
    def error_estimation(
            self, base_distribution, pull_tar, solve_params, full_output=False):
        if solve_params['qtype'] != 0:
            raise AttributeError(
                "The Gradient Stability validator is defined only for " + \
                "Monte Carlo quadrature rules")
        tm = pull_tar.transport_map
        # Compute mean and variance of the gradient
        nsamps = int(np.ceil( self.n_grad_samps * solve_params['qparams'] ))
        x, w = base_distribution.quadrature(
            qtype=solve_params['qtype'], qparams=nsamps)
        scatter_tuple = (['x'], [x])
        if not self.fungrad:
            grad = - mpi_map(
                "grad_a_log_pdf", obj=pull_tar, scatter_tuple=scatter_tuple,
                mpi_pool=solve_params.get('mpi_pool'))
        else:
            _, grad = - mpi_map(
                "tuple_grad_a_log_pdf", obj=pull_tar, scatter_tuple=scatter_tuple,
                mpi_pool=solve_params.get('mpi_pool'))
        if solve_params['regularization'] is not None:
            if solve_params['regularization']['type'] == 'L2':
                grad += solve_params['regularization']['alpha'] * 2. * \
                        (tm.coeffs - tm.get_identity_coeffs())
        # Use randomized resampling to avoid being forced to use n_grad_samps > tm.n_coeffs
        if self.n_bootstrap is None:
            avg_grads = np.mean( grad.reshape((
                solve_params['qparams'], self.n_grad_samps, tm.n_coeffs)), axis=0 )
        else:
            avg_grads = np.zeros((self.n_bootstrap, tm.n_coeffs))
            for i in range(self.n_bootstrap):
                idxs = npr.choice(nsamps, size=solve_params['qparams'], replace=True)
                avg_grads[i,:] = np.mean( grad[idxs,:], axis=0 )
        mean = np.mean(avg_grads, axis=0)
        cntr_avg_grads = avg_grads - mean
        cov = np.dot(cntr_avg_grads.T, cntr_avg_grads) / float(solve_params['qparams']-1)
        L = scila.cholesky(cov, lower=True)
        cstar = stats.chi2(tm.n_coeffs).ppf(self.alpha)
        # Solve maximization problem
        A = - cstar/solve_params['qparams'] * np.dot(L.T,L)
        b = - 2. * np.sqrt(cstar/solve_params['qparams']) * np.dot(mean, L)
        c = - np.dot(mean, mean)
        def f(x, A, b, c):
            return np.dot(x,np.dot(A,x)) + np.dot(b, x) + c
        def jac(x, A, b, c):
            return 2. * np.dot(A,x) + b
        def c1(x):
            return 1 - np.dot(x,x)
        def c1jac(x):
            return - 2 * x
        cons = ({'type': 'ineq', 'fun': c1, 'jac': c1jac})
        res = sciopt.minimize(
            f, np.zeros(tm.n_coeffs), method='SLSQP', jac=jac, 
            args=(A, b, c), constraints=cons, tol=1e-12)
        xstar = res['x']
        # Compute err as sqrt(f(x))
        err = np.sqrt(- f(xstar, A, b, c))
        if full_output:
            # Refinement parameters
            ref_params = {
                'mean':mean,
                'L': L,
                'cstar': cstar,
                'xstar': xstar
            }
            # Prune parameter as 1/variance of directional gradient
            var_coeffs = np.diag(cov)
            prune_params = (1/var_coeffs)/max(1/var_coeffs)
            return err, ref_params, prune_params
        else:
            return err

    def refinement(self, base_distribution, pull_tar,
                   qtype, qparams, ref_params):
        if qtype != 0:
            raise AttributeError(
                "The Gradient Stability validator is defined only for " + \
                "Monte Carlo quadrature rules")
        # Load parameters
        mu = ref_params['mean']
        L = ref_params['L']
        cstar = ref_params['cstar']
        xstar = ref_params['xstar']
        # Compute refinement
        a = np.dot(mu,mu) - self.eps_sp**2
        if a <= 0:
            Lx = np.dot(L, xstar)
            b = 2. * np.sqrt(cstar) * np.dot(mu, np.dot(L, xstar))
            c = cstar * np.dot(Lx.T, Lx)
            z = (-b + np.sqrt(b**2-4*a*c))/ 2. / a
            q = min( int(np.ceil(z**2)),
                     self.lmb_max * qparams )
        else:
            q = int( np.ceil(self.lmb_def * qparams) )
        return q, q
        
class GradientStabilityKLMinimizationValidator(KLMinimizationValidator):
    def __init__(self, eps_sp, cost_function, max_cost,
                 max_nsamps=np.inf, stop_on_fcast=False,
                 n_grad_samps=10, n_gap_resampling=10,
                 beta=0.95, gamma=0.05, lmb_def=2, lmb_max=10,
                 fungrad=False):
        self.n_grad_samps = n_grad_samps
        self.n_gap_resampling = n_gap_resampling
        self.beta = beta
        self.gamma = gamma
        self.lmb_def = lmb_def
        self.lmb_max = lmb_max
        self.fungrad = fungrad
        super(GradientStabilityKLMinimizationValidator,
              self).__init__(eps_sp, cost_function, max_cost, max_nsamps, stop_on_fcast)

    def _cv(self, N, n, b, g):
        n = self.n_grad_samps
        b = self.beta
        g = self.gamma
        return (n-1) * stats.chi2(N, loc=float(N)/float(n)).ppf(b) / \
            stats.chi2(n-N).ppf(1-g)
        
    def error_estimation(
            self, base_distribution, pull_tar, solve_params, full_output=False):
        if solve_params['qtype'] != 0:
            raise AttributeError(
                "The Gradient Stability validator is defined only for " + \
                "Monte Carlo quadrature rules")
        tm = pull_tar.transport_map
        # Compute mean and variance of the gradient
        nsamps = int(np.ceil( self.n_grad_samps * solve_params['qparams'] ))
        x, w = base_distribution.quadrature(
            qtype=solve_params['qtype'], qparams=nsamps)
        scatter_tuple = (['x'], [x])
        if not self.fungrad:
            grad = - mpi_map(
                "grad_a_log_pdf", obj=pull_tar, scatter_tuple=scatter_tuple,
                mpi_pool=solve_params.get('mpi_pool'))
        else:
            _, grad = - mpi_map(
                "tuple_grad_a_log_pdf", obj=pull_tar, scatter_tuple=scatter_tuple,
                mpi_pool=solve_params.get('mpi_pool'))
        if solve_params['regularization'] is not None:
            if solve_params['regularization']['type'] == 'L2':
                grad += solve_params['regularization']['alpha'] * 2. * \
                        (tm.coeffs - tm.get_identity_coeffs())
        # Use randomized resampling to avoid being forced to use n_grad_samps > tm.n_coeffs
        n = tm.n_coeffs + self.n_gap_resampling
        avg_grads = np.zeros((n, tm.n_coeffs))
        for i in range(n):
            idxs = npr.choice(nsamps, size=solve_params['qparams'], replace=False)
            avg_grads[i,:] = np.mean( grad[idxs,:], axis=0 )
        mean = np.mean(avg_grads, axis=0)
        cntr_avg_grads = avg_grads - mean
        cov = np.dot(cntr_avg_grads.T, cntr_avg_grads) / float(n-1)
        L = scila.cholesky(cov, lower=True)
        cstar = self._cv(tm.n_coeffs, n, self.beta, self.gamma)
        # Solve maximization problem
        A = - cstar/solve_params['qparams'] * np.dot(L.T,L)
        b = - 2. * np.sqrt(cstar/solve_params['qparams']) * np.dot(mean, L)
        c = - np.dot(mean, mean)
        def f(x, A, b, c):
            return np.dot(x,np.dot(A,x)) + np.dot(b, x) + c
        def jac(x, A, b, c):
            return 2. * np.dot(A,x) + b
        def c1(x):
            return 1 - np.dot(x,x)
        def c1jac(x):
            return - 2 * x
        cons = ({'type': 'ineq', 'fun': c1, 'jac': c1jac})
        res = sciopt.minimize(
            f, np.zeros(tm.n_coeffs), method='SLSQP', jac=jac, 
            args=(A, b, c), constraints=cons, tol=1e-12)
        xstar = res['x']
        # Compute err as sqrt(f(x))
        err = np.sqrt(f(xstar, A, b, c))
        if full_output:
            # Refinement parameters
            ref_params = {
                'mean':mean,
                'L': L,
                'cstar': cstar,
                'xstar': xstar
            }
            # Compute prune parameters as 1/ variance of gradient
            var_coeffs = np.diag(cov)
            prune_params = (1/var_coeffs)/max(1/var_coeffs)
            return err, ref_params, prune_params
        else:
            return err

    def refinement(self, base_distribution, pull_tar,
                   qtype, qparams, ref_params):
        if qtype != 0:
            raise AttributeError(
                "The Gradient Stability validator is defined only for " + \
                "Monte Carlo quadrature rules")
        # Load parameters
        mu = ref_params['mean']
        L = ref_params['L']
        cstar = ref_params['cstar']
        xstar = ref_params['xstar']
        # Compute refinement
        a = np.dot(mu,mu) - self.eps_sp**2
        if a <= 0:
            Lx = np.dot(L, xstar)
            b = 2. * np.sqrt(cstar) * np.dot(mu, np.dot(L, xstar))
            c = cstar * np.dot(Lx.T, Lx)
            z = (-b + np.sqrt(b**2-4*a*c))/ 2. / a
            q = int(np.ceil(z**2))
        else:
            q = int( np.ceil(self.lmb_def * qparams) )
        return q, q




#!/usr/bin/env python

#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import sys, getopt
import numpy as np
import logging

import TransportMaps.Distributions.Examples.BiochemicalOxigenDemand as BOD

def usage():
    str_usage = """
DataGeneration.py --output=FNAME
  [--numy=N --timey=N --sigma=N]
    """
    print(str_usage)

def full_usage():
    usage()

argv = sys.argv[1:]
numy = 0
timey = 0.
sigma = np.sqrt(1e-3)
OUT_FNAME = None
try:
    opts, args = getopt.getopt(argv,"h", [
        "output=",
        "numy=", "timey=", "sigma="])
except getopt.GetoptError:
    full_usage()
    sys.exit(2)
for opt, arg in opts:
    if opt == '-h':
        full_usage()
        sys.exit()
    elif opt in ("--output"):
        OUT_FNAME = arg
    elif opt in ("--numy"):
        numy = int(arg)
    elif opt in ("--timey"):
        timey = float(arg)
    elif opt in ("--sigma"):
        sigma = float(arg)
if None in [OUT_FNAME]:
    logging.warn("The density won't be stored")
    
d = BOD.BODjoint(numy, timey, sigma)

if OUT_FNAME is not None:
    d.store(OUT_FNAME)

#####
# PLOTTING
import TransportMaps.Diagnostics as DIAG
DIAG.plotAlignedConditionals(d, range_vec=[-5,5])


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import sys
import os
import os.path
import dill
import numpy as np

from . import AvailableOptions as AO
from .ScriptBase import Script

from TransportMaps.External import H5PY_SUPPORT, PLOT_SUPPORT
import TransportMaps.Distributions as DIST
import TransportMaps.Maps as MAPS
from TransportMaps.CLI.PostprocessBase import Postprocess

if PLOT_SUPPORT:
    import matplotlib.pyplot as plt
    
if H5PY_SUPPORT:
    import h5py

__all__ = ['AdaptivityPostprocess']

class AdaptivityPostprocess(Postprocess):

    cmd_usage_str = "Usage: tmap-adaptivity-postprocess "
    opts_usage_str = """  [--adapt-step=N --var-diag-conv]
"""
    usage_str = cmd_usage_str + Postprocess.opts_usage_str + opts_usage_str

    docs_descr_str = """DESCRIPTION
Given a file (--input) storing the transport maps pushing forward a base distribution
to a target distribution constructed through an adaptivity algorithm,
provides a number of diagnositic routines.
All files involved are stored and loaded using the python package dill and
an extra file OUTPUT.hdf5 is created to store big datasets in the hdf5 format.
In the following default values are shown in brackets."""
    docs_options_adapt_str = """OPTIONS - Adaptivity:
  --adapt-step=N          Run postprocessing analysis on the N-th map of the adaptivity
  --var-diag-conv         Plot the convergence in variance diagnostic vs. number
                            of coefficients
"""
    docs_options_str = Postprocess.docs_options_str + docs_options_adapt_str
    docs_str = docs_descr_str + docs_options_str

    def usage(self):
        print(AdaptivityPostprocess.usage_str)

    def description(self):
        print(AdaptivityPostprocess.docs_str)

    @property
    def long_options(self):
        return super(AdaptivityPostprocess, self).long_options + \
            [
                "adapt-step=", "var-diag-conv",
            ]

    def load_opts(self, opts):
        super(AdaptivityPostprocess, self).load_opts(opts)
        for opt, arg in opts:
            if opt == "--adapt-step":
                self.ADAPT_STEP = int(arg)
            elif opt == "--var-diag-conv":
                self.ADAPT_VAR_DIAG_CONV = True

    def __init__(self, argv):
        self.ADAPT_STEP = -1
        self.ADAPT_VAR_DIAG_CONV = False
        super(AdaptivityPostprocess, self).__init__(argv)

    def load(self):
        self.h5_file = None
        
        # Load data
        with open(self.INPUT, 'rb') as in_stream:
            self.stg = dill.load(in_stream)

        # Restore data
        self.base_distribution = self.stg.base_distribution
        self.target_distribution = self.stg.target_distribution
        try:
            self.builder = self.stg.builder
        except AttributeError:
            raise AttributeError(
                "The input file does not contain and adaptivity builder.")
        try:
            tm = self.builder.transport_map_list[self.ADAPT_STEP]
        except AttributeError:
            raise AttributeError(
                "The builder is not adaptive.")
        except IndexError:
            nmaps = len(self.builder.transport_map_list)
            raise IndexError(
                "Only %d maps are available. Required %d > %d." % (
                    nmaps, self.ADAPT_STEP, nmaps))
        self.tmap = MAPS.CompositeMap( self.stg.precond_map, tm )
        
        self.approx_base_distribution = DIST.PullBackTransportMapDistribution(
            self.tmap, self.target_distribution)
        self.approx_target_distribution = DIST.PushForwardTransportMapDistribution(
            self.tmap, self.base_distribution)
        self.dim = self.base_distribution.dim

        if self.ADAPT_STEP == -1:
            self.ADAPT_STEP = len(self.builder.transport_map_list)-1

        # Load output (dill file) if any
        if not os.path.exists(self.OUTPUT):
            self.postproc_data = {}
            with open(self.OUTPUT, 'wb') as out_stream:
                dill.dump(self.postproc_data, out_stream)
        with open(self.OUTPUT, 'rb') as in_stream:
            self.postproc_data = dill.load(in_stream)
        if not hasattr(self.postproc_data, 'adapt-steps'):
            self.postproc_data['adapt-steps'] = [{}] * len(self.builder.transport_map_list)
        self.postproc_data['adapt-steps'] += \
            [{}] * (len(self.builder.transport_map_list) - \
                    len(self.postproc_data['adapt-steps']))
        self.postproc_root = self.postproc_data['adapt-steps'][self.ADAPT_STEP]
        
        # Load output (hdf5 file) if any
        self.h5_file = h5py.File(self.OUTPUT + '.hdf5', 'a')
        ROOT_NAME = "adapt-step-%d" % self.ADAPT_STEP
        if ROOT_NAME not in self.h5_file:
            self.h5_file.create_group(ROOT_NAME)
        self.h5_root = self.h5_file[ROOT_NAME]

    def variance_diagnostic_convergence(self, mpi_pool):
        if self.ADAPT_VAR_DIAG_CONV:
            n_coeffs_list = [ self.builder.transport_map_list[i].n_coeffs
                              for i in range(len(self.builder.variance_diagnostic_list)) ]
            plt.figure()
            plt.semilogy(n_coeffs_list, self.builder.variance_diagnostic_list, '--o')
            plt.show(False)

    def run(self, mpi_pool):
        self.variance_diagnostic_convergence(mpi_pool)
        super(AdaptivityPostprocess, self).run(mpi_pool)


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

from collections import deque
import numpy as np
import numpy.linalg as npla
import numpy.random as npr

import TransportMaps as TM

__all__ = ['adaptive_randomized_range_finder',
           'randomized_direct_eigenvalue_decomposition',
           'randomized_nystrom_eigenvalue_decomposition']

nax = np.newaxis

def adaptive_randomized_range_finder(
        action, dim_in, dim_out, eps, r, rng=npr.randn, kwargs={},
        power_n=0, action_transpose=None):
    r""" Adaptive randomized range finder with subspace iterations

    .. see:: Algorithms 4.2 and 4.4 in Halko, N., Martinsson, P., & Tropp, J. (2011). Finding structure with randomness: Probabilistic algorithms for constructing approximate matrix decompositions. SIAM Review, 53(2), 217–288. Retrieved from http://epubs.siam.org/doi/abs/10.1137/090771806
    """
    Omega = rng(dim_in*r).reshape((dim_in, r))
    Y = action(Omega, **kwargs)
    for k in range(power_n):
        Y, R = npla.qr(Y)
        Y = action_transpose(Y, **kwargs)
        Y, R = npla.qr(Y)
        Y = action(Y, **kwargs)
    Y = [Y[:,[i]] for i in range(r)]
    j = 0
    Q = np.zeros((dim_out,0))
    nrm_list = deque([npla.norm(y) for y in Y])
    tar_eps = eps / (10. * np.sqrt(2/np.pi))
    TM.logger.info(
        "Randomized adaptive range finder: it=%d - " % j + \
        "eps=%e (target %e)" % (max(nrm_list), tar_eps))
    while max(nrm_list) > tar_eps or j == 0:
        Y[j] -= np.dot(Q, np.dot(Q.T, Y[j]))
        q = Y[j] / npla.norm(Y[j])
        Q = np.hstack((Q, q))
        # New direction
        w = rng(dim_in).reshape((dim_in,1))
        Aw = action(w, **kwargs)
        for k in range(power_n):
            Aw, R = npla.qr(Aw)
            Aw = action_transpose(Aw, **kwargs)
            Aw, R = npla.qr(Aw)
            Aw = action(Aw, **kwargs)
        Y.append( Aw - np.dot(Q, np.dot(Q.T, Aw)) )
        for i in range(j+1, j+r+1):
            Y[i] -= Q[:,[j]] * np.dot(Q[:,j], Y[i])
        # Update norms
        nrm_list.popleft()
        nrm_list.append( npla.norm(Y[-1]) )
        j += 1
        TM.logger.info(
            "Randomized adaptive range finder: it=%d - " % j + \
            "eps=%e (target %e)" % (max(nrm_list), tar_eps))
    return Q

def randomized_direct_eigenvalue_decomposition(
        action, dim, eps, r, rng=npr.randn, kwargs={},
        power_n=0, range_finder=adaptive_randomized_range_finder,
        min_eigval=1e-12):
    Q = range_finder(
        action, dim, dim, eps, r, rng=rng, kwargs=kwargs,
        power_n=power_n, action_transpose=action)
    AQ = action(Q, **kwargs)
    B = np.dot(Q.T, AQ)
    D, V = npla.eig(B)
    # Remove imaginary part
    D = D.real
    V = V.real
    # Remove eigenvalues smaller than min_eigval
    idxs = np.abs(D) > min_eigval
    TM.logger.info("Randomized direct eigenvalue decomposition - Truncation dimension: %d" % len(idxs))
    D = D[idxs]
    V = V[:,idxs]
    U = np.dot(Q, V)
    return D, U

def randomized_nystrom_eigenvalue_decomposition(
        action, dim, eps, r, rng=npr.randn, kwargs={},
        power_n=0, range_finder=adaptive_randomized_range_finder,
        min_eigval=1e-12):
    Q = range_finder(
        action, dim, dim, eps, r, rng=rng, kwargs=kwargs,
        power_n=power_n, action_transpose=action)
    B1 = action(Q, **kwargs)
    B2 = np.dot(Q.T, B1)
    L = npla.cholesky(B2)
    F = npla.solve_triangular(L, B1.T).T
    U,S,V = npla.svd(F)
    D = S**2.
    # Remove eigenvalues smaller than min_eigval
    idxs = np.abs(D) > min_eigval
    TM.logger.info("Randomized direct eigenvalue decomposition - Truncation dimension: %d" % len(idxs))
    D = D[idxs]
    U = U[:,idxs]
    return D, U


#!/usr/bin/env python

#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import sys, getopt
import numpy as np
import dill
import h5py
import pyhmc
import matplotlib.pyplot as plt

nax = np.newaxis

argv = sys.argv[1:]
IN_FNAME = None
OUT_FNAME = None
NSAMPS = None
PLOTTING = False
NBURNIN = 0

try:
    opts, args = getopt.getopt(argv,"hn:",["output=", "dist=", "plotting","burnin="])
except getopt.GetoptError:
    full_usage()
    sys.exit(2)
for opt, arg in opts:
    if opt == '-h':
        full_usage()
        sys.exit()
    elif opt == "-n":
        NSAMPS = int(arg)
    elif opt == "--dist":
        IN_FNAME = arg
    elif opt == "--output":
        OUT_FNAME = arg
    elif opt == "--plotting":
        PLOTTING = True
    elif opt == '--burnin':
        NBURNIN = int(arg)
if None in [IN_FNAME,OUT_FNAME]:
    full_usage()
    sys.exit(3)

with open(IN_FNAME,'rb') as istr:
    d = dill.load(istr)

f = h5py.File(OUT_FNAME,'a')
DSET_NAME = 'hmc'
if DSET_NAME not in f:
    f.create_dataset(DSET_NAME, (0,d.dim), maxshape=(None,d.dim), dtype='d')
loaded_samps = f[DSET_NAME]            

def logprob(x, d):
    return d.log_pdf(x[nax,:])[0], d.grad_x_log_pdf(x[nax,:])[0,:]

nold = loaded_samps.shape[0]
nnew = NSAMPS - nold
if nold == 0:
    x0 = np.random.randn(d.dim)
else:
    x0 = loaded_samps[-1,:]

if nnew > 0:
    samples = pyhmc.hmc(logprob, x0=x0, args=(d,), n_samples=nnew, display=True)
    loaded_samps.resize(NSAMPS, axis=0)
    loaded_samps[nold:,:] = samples

samples = loaded_samps[:,:]
    
f.close()

if PLOTTING:
    # MU
    mu = d.mu.evaluate(samples[NBURNIN:,0])
    plt.figure()
    plt.hist(mu, bins=50, edgecolor='k', facecolor='w', normed=True)
    plt.xlabel(r"$\mu$")
    # SIGMA
    sigma = d.sigma.evaluate(samples[NBURNIN:,1])
    tau = 1/sigma**2
    plt.figure()
    plt.hist(tau, bins=50, edgecolor='k', facecolor='w', normed=True)
    plt.xlabel(r"$\tau$")
    # PHI
    phi = d.phi.evaluate(samples[NBURNIN:,2])
    plt.figure()
    plt.hist(phi, bins=50, edgecolor='k', facecolor='w', normed=True)
    plt.xlabel(r"$\phi$")
    plt.show(False)

    # Chains
    plt.figure()
    plt.plot(mu)
    plt.xlabel(r"$\mu$")
    plt.figure()
    plt.plot(sigma)
    plt.xlabel(r"$\sigma$")
    plt.figure()
    plt.plot(phi)
    plt.xlabel(r"$\phi$")

    # ESS
    import TransportMaps.Samplers as SMPL
    ess = SMPL.ess(samples[NBURNIN:,:3], plotting=True, plot_lag=100000)
    plt.show(False)

    print("Effective sample size: %d" % ess)

# from IPython import embed
# embed()




#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import numpy as np
import scipy.linalg as scila
import scipy.integrate as sciint

from TransportMaps.Distributions.FrozenDistributions import GaussianDistribution
from TransportMaps.Distributions.Decomposable.LinearGaussianSequentialInferenceDistributions import \
    LinearGaussianAR1TransitionDistribution
from TransportMaps.Likelihoods.LikelihoodBase import \
    AdditiveLinearGaussianLogLikelihood

__all__ = ['Prior', 'Transition', 'LogLikelihood', 'generate_data']

re = 6367253.
Omega = 0.2625161
dt = 0.5

class Prior(GaussianDistribution):
    def __init__(self):
        sig2_eps = (0.02 * np.pi / 180)**2
        Q0 = np.zeros((6,6))
        Q0[0,0] = Q0[1,1] = (1000./re)**2
        Q0[2,2] = (0.1 * np.pi/180)**2
        Q0[3,3] = Q0[4,4] = Q0[5,5] = sig2_eps
        # Init
        super(Prior,self).__init__(np.zeros(6), Q0)

class Transition(LinearGaussianAR1TransitionDistribution):
    def __init__(self):
        theta = np.pi / 4
        Ox = Omega * np.cos(theta)
        Oz = Omega * np.sin(theta)
        F = np.zeros((6,6))
        F[0,1] = Oz; F[1,0] = -Oz; F[1,2] = Ox; F[2,1] = -Ox;
        F[0,3] = F[1,4] = F[2,5] = 1.
        G = np.zeros((6,6))
        G[3,3] = G[4,4] = G[5,5] = 1.
        sig2_eps = (0.02 * np.pi / 180)**2
        W = np.eye(6) * sig2_eps
        # Integrate dynamics to obtain finite-difference equations
        Phi = scila.expm(F*dt)
        GWG = np.dot(G, np.dot(W,G))
        def f(q,t,F,GWG):
            Q = q.reshape((6,6))
            rhs = np.dot(F,Q) + np.dot(Q,F.T) + GWG
            return rhs.flatten()
        q0 = np.zeros(6*6)
        T = [0, dt]
        sol = sciint.odeint(f, q0, T, args=(F, GWG))
        Q = sol[1].reshape((6,6))
        # Init
        super(Transition,self).__init__(np.zeros(6), Phi, np.zeros(6), Q)

class LogLikelihood(AdditiveLinearGaussianLogLikelihood):
    def __init__(self, y):
        sig2_xy = (1000./re)**2
        sig2_z = (1./60. * np.pi/180)**2
        H = np.zeros((3,6))
        H[0,0] = H[1,1] = H[2,0] = H[2,2] = 1
        R = np.zeros((3,3))
        R[0,0] = R[1,1] = sig2_xy
        R[2,2] = sig2_z
        super(LogLikelihood,self).__init__(y, np.zeros(3), H, np.zeros(3), R)

def generate_data(nsteps):
    prior = Prior()
    trans = Transition()
    ll = LogLikelihood(np.zeros(3)) # Just use the constructor to get handle on params
    # Generate dynamics
    T = np.zeros(nsteps+1)
    Z = np.zeros((nsteps+1, 6))
    Z[0,:] = prior.rvs(1)[0,:]
    for i in range(nsteps):
        T[i+1] = T[i] + dt
        Z[i+1,:] = trans.T.evaluate(Z[[i],:])[0,:] + trans.pi.rvs(1)[0,:]
    # Generate observations
    Y = []
    for i in range(nsteps+1):
        Y.append( ll.T.evaluate(Z[[i],:])[0,:] + ll.pi.rvs(1)[0,:] )
    return (T,Z,Y)


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

from .AdaptivityBase import Builder
from TransportMaps.Routines import L2_misfit

__all__ = ['L2RegressionBuilder',
           'ToleranceSequentialL2RegressionBuilder']

class L2RegressionBuilder(Builder):
    r""" Basis builder through :math:`\mathcal{L}^2` regression

    Given a map :math:`M`, fit a supplied parametric transport map :math:`T`
    through the solution of the :math:`\mathcal{L}^2` regression problem

    .. math::

       \arg\min_{\bf a}\left\Vert M - T \right\Vert_{\mathcal{L}^2} + \alpha \Vert {\bf a} \Vert_x

    where :math:`\alpha\Vert{\bf a}\Vert_x` is a regularization term, with
    respect to one of the available norms.

    Args:
      regression_params (dict): dictionary of regression parameters
    """
    def __init__(self, regression_params):
        self.regression_params = regression_params
        super(L2RegressionBuilder, self).__init__()
        
    def solve(self, transport_map, target_map, **extra_reg_params):
        params = dict(self.regression_params, **extra_reg_params)
        log_list = transport_map.regression(
            target_map, **params)
        return transport_map, log_list

class ToleranceSequentialL2RegressionBuilder(L2RegressionBuilder):
    def __init__(self, eps, regression_params_list, monitor_params):
        self.eps = eps
        self.regression_params_list = regression_params_list
        self.monitor_params = monitor_params
        super(ToleranceSequentialL2RegressionBuilder,
              self).__init__(self.regression_params_list[0])
    
    def solve(self, transport_map_list, target_map, **extra_reg_params):
        tm = None
        x0 = None
        for transport_map, self.regression_params in zip(
                transport_map_list, self.regression_params_list):
            if tm is not None:
                for c1, c2 in zip(tm.approx_list, self.transport_map.approx_list):
                    # Constant part
                    for i1, midx1 in enumerate(c1.c.multi_idxs):
                        for i2, midx2 in enumerate(c2.c.multi_idxs):
                            if midx1 == midx2:
                                break
                        c2.c.coeffs[i2] = c1.c.coeffs[i1]
                    # Integrated part
                    for i1, midx1 in enumerate(c1.h.multi_idxs):
                        for i2, midx2 in enumerate(c2.h.multi_idxs):
                            if midx1 == midx2:
                                break
                        c2.h.coeffs[i2] = c1.h.coeffs[i1]
                self.regression_params['x0'] = transport_map.coeffs
            tm_new, log_list = super(
                ToleranceSequentialL2RegressionBuilder,
                self).solve(transport_map, target_map, **extra_reg_params)
            if not log_list[-1]['success']:
                if tm is None: # If no map is available return the target map
                    return target_map, log_list
                else:
                    break
            else:
                tm = transport_map
            # Check error
            err = L2_misfit(target_map, tm, **self.monitor_params)
            if err < eps:
                break
        return transport_map, log_list
            


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import unittest
import numpy as np
import numpy.random as npr

from TransportMaps import MPI_SUPPORT

class L2_minimization(unittest.TestCase):

    def setUp(self):
        npr.seed(1)
        self.reg = {'type': 'L2',
                    'alpha': 1e-7}
        self.tol = 1e-3
        self.dim = 2
        self.order = 3
        self.span = 'total'
        self.batch_size_list = [(None,None,None),(None,None,5)]

    def test_L2_minimization(self):
        import TransportMaps as TM
        import TransportMaps.Distributions as DIST

        # Approximate
        qtype = 3
        qparams = [10] * self.dim
        d = DIST.StandardNormalDistribution(self.dim)

        log = self.tm_approx.regression(self.target, d=d, qtype=qtype, qparams=qparams,
                                        regularization=self.reg, tol=self.tol,
                                        batch_size_list=self.batch_size_list,
                                        mpi_pool_list=[self.mpi_pool]*self.dim)

        # Check L2 accuracy
        (x,w) = d.quadrature(qtype,qparams)
        t1 = self.target.evaluate(x)
        t2 = self.tm_approx.evaluate(x)
        misfit = np.sqrt( np.sum((t1-t2)**2, axis=1) )
        l2_misfit = np.dot(misfit, w)
        assertion = l2_misfit < 10 * self.dim * self.tol
        if not assertion:
            print("Misfit > 10 * d * tol : %e > %e" % (l2_misfit, 10 * self.dim * self.tol))
        self.assertTrue( assertion )

#
# Monotone types
#
class IntExp_L2_min(L2_minimization):
    def setUp(self):
        import TransportMaps as TM
        super(IntExp_L2_min,self).setUp()
        # Generate a random integrated exponential map
        self.target = TM.Default_IsotropicIntegratedExponentialTriangularTransportMap(
            self.dim, self.order, self.span)
        self.target.coeffs = np.random.randn( self.target.n_coeffs ) / 10. 
        # Generate approximation
        self.tm_approx = TM.Default_IsotropicIntegratedExponentialTriangularTransportMap(
            self.dim, self.order, self.span)
        self.ders = 2
class IntSq_L2_min(L2_minimization):
    def setUp(self):
        import TransportMaps as TM
        super(IntSq_L2_min,self).setUp()
        # Generate a random integrated exponential map
        self.target = TM.Default_IsotropicIntegratedSquaredTriangularTransportMap(
            self.dim, self.order, self.span)
        coeffs = self.target.get_identity_coeffs()
        self.target.coeffs = coeffs + np.random.randn( self.target.n_coeffs ) / 10. 
        # Generate approximation
        self.tm_approx = TM.Default_IsotropicIntegratedSquaredTriangularTransportMap(
            self.dim, self.order, self.span)
        self.ders = 2
class LinSpan_L2_min(L2_minimization):
    def setUp(self):
        import TransportMaps as TM
        super(LinSpan_L2_min,self).setUp()
        # Generate a random integrated exponential map
        self.target = TM.Default_IsotropicLinearSpanTriangularTransportMap(
            self.dim, self.order, self.span)
        self.target.coeffs = np.random.randn( self.target.n_coeffs ) / 10. 
        # Generate approximation
        self.tm_approx = TM.Default_IsotropicLinearSpanTriangularTransportMap(
            self.dim, self.order, self.span)
        self.ders = 1

#
# Serial/Parallel    
#
class Serial_IntExp_L2_min(IntExp_L2_min):
    def setUp(self):
        super(Serial_IntExp_L2_min, self).setUp()
        self.mpi_pool = None
class Serial_IntSq_L2_min(IntSq_L2_min):
    def setUp(self):
        super(Serial_IntSq_L2_min, self).setUp()
        self.mpi_pool = None
class Serial_LinSpan_L2_min(LinSpan_L2_min):
    def setUp(self):
        super(Serial_LinSpan_L2_min, self).setUp()
        self.mpi_pool = None
class Parallel_IntExp_L2_min(IntExp_L2_min):
    def setUp(self):
        super(Parallel_IntExp_L2_min, self).setUp()
        import TransportMaps as TM
        import_set = set([ (None, 'numpy', 'np') ])
        self.mpi_pool = TM.get_mpi_pool()
        self.mpi_pool.start(2)
        self.mpi_pool.mod_import(import_set)
    def tearDown(self):
        import time
        self.mpi_pool.stop()
        time.sleep(0.2)
class Parallel_IntSq_L2_min(IntSq_L2_min):
    def setUp(self):
        super(Parallel_IntSq_L2_min, self).setUp()
        import TransportMaps as TM
        import_set = set([ (None, 'numpy', 'np') ])
        self.mpi_pool = TM.get_mpi_pool()
        self.mpi_pool.start(2)
        self.mpi_pool.mod_import(import_set)
    def tearDown(self):
        import time
        self.mpi_pool.stop()
        time.sleep(0.2)
class Parallel_LinSpan_L2_min(LinSpan_L2_min):
    def setUp(self):
        super(Parallel_LinSpan_L2_min, self).setUp()
        import TransportMaps as TM
        import_set = set([ (None, 'numpy', 'np') ])
        self.mpi_pool = TM.get_mpi_pool()
        self.mpi_pool.start(2)
        self.mpi_pool.mod_import(import_set)
    def tearDown(self):
        import time
        self.mpi_pool.stop()
        time.sleep(0.2)
        
def build_suite(ttype='all'):
    suite_se_ie_l2_min = unittest.TestLoader().loadTestsFromTestCase(
        Serial_IntExp_L2_min )
    suite_se_is_l2_min = unittest.TestLoader().loadTestsFromTestCase(
        Serial_IntSq_L2_min )
    suite_se_ls_l2_min = unittest.TestLoader().loadTestsFromTestCase(
        Serial_LinSpan_L2_min )
    suite_pa_ie_l2_min = unittest.TestLoader().loadTestsFromTestCase(
        Parallel_IntExp_L2_min )
    suite_pa_is_l2_min = unittest.TestLoader().loadTestsFromTestCase(
        Parallel_IntSq_L2_min )
    suite_pa_ls_l2_min = unittest.TestLoader().loadTestsFromTestCase(
        Parallel_LinSpan_L2_min )

    # GROUP SUITES
    # Serial
    suites_list = []
    if ttype in ['all','serial']:
        suites_list += [
            suite_se_ie_l2_min, suite_se_is_l2_min, suite_se_ls_l2_min
        ]
    if ttype in ['all','parallel'] and MPI_SUPPORT:
        # Parallel
        suites_list += [
            suite_pa_ie_l2_min, suite_pa_is_l2_min, suite_pa_ls_l2_min
        ]

    all_suites = unittest.TestSuite( suites_list )
    return all_suites

def run_tests(ttype='all'):
    all_suites = build_suite(ttype)
    # RUN
    unittest.TextTestRunner(verbosity=2).run(all_suites)
    
if __name__ == '__main__':
    run_tests()
